#!/bin/sh
# $Id: common.sh,v 1.102 2014/02/13 15:59:01 ca Exp $
# Copyright (c) 2004-2006, 2011 Sendmail, Inc. and its suppliers.
#	All rights reserved.
#
# By using this file, you agree to the terms and conditions set
# forth in the LICENSE file which can be found at the top level of
# the sendmail distribution.
#
# ----------------------------------------
# set common variables for test programs
# needs: SD
# NOTE: this also sources smar/common.sh!
# ----------------------------------------

test -s stop && exit 1

SCRIPTS=`cd ${SD} && pwd`
THISDIR=`pwd`

# paths etc for various (test) programs
DEFEDB=${MTA_CDB_BASE_DIR}defedb
IBDBDIR=${MTA_CDB_BASE_DIR}'ibdb/ibdb'
IBDB='ibd00*'
SMTPCSOCK=${MTA_CDB_BASE_DIR}qmsmtpc/qmsmtpc
SMARSOCK=${MTA_CDB_BASE_DIR}"qmsmar/qmsmar"

# path to CDB for lmtp sink (smtps3)
LMTPDIR="lmtpdir/"
SMTPDIR="smtpdir/"

# some filenames for the sinks
B1=${LMTPDIR}1/S000000000000000100
B2=${LMTPDIR}2/S000000000000000200
B3=${LMTPDIR}3/S000000000000000300
S1=${SMTPDIR}1/S000000000000000100
S2=${SMTPDIR}2/S000000000000000200
S3=${SMTPDIR}3/S000000000000000300
S4=${SMTPDIR}4/S000000000000000400

# port numbers to use
SNKPORT=${MTA_SNKPORT:-8753}
SRVPORT=${MTA_SRVPORT:-1579}

# some logfile names
SRVL="sink.log"
LMTPL="lmtp.log"
CLTL="source.log"
CLTL2="source2.log"
CLTL3="source3.log"
AML="a0.log"
QML="q0.log"
PML="m0.log"
SCL="c0.log"
SSL="s0.log"
# See V=1 below!
AML1="a1.log"
QML1="q1.log"
SCL1="c1.log"
SSL1="s1.log"
MCPLOGS="smtps.log smtpc.log qmgr.log smar.log MTA.log MSA.log"

# host on which to listen/to which to send
H=localhost
PIDS=pids
PID2=pids2
PMIDS=pmids
STE="../statethreads/examples"
EDBR=${EDBR:-../checks/t-edbr-0}
export STE CLTL EDBR H PIDS

#
SMXCNF=smx.conf
MCPCNF=${SMXCNF}
#MCPCNF=mcp.conf
CONF2=smx2.conf

CAS=CA.certs.pem

# do not run as root
../libcheck/noroot || exit 1

# get USER
. ${SD}/../check2/user.sh

#
AWK=awk
T=/usr/bin/nawk
[ -x $T ] && AWK=$T
CAT=cat
DIFF=diff
EGREP=egrep
XPG=/usr/xpg4/bin
GREP=grep
T=${XPG}/grep
[ -x $T ] && GREP=$T
TAIL=tail
T=${XPG}/tail
[ -x $T ] && TAIL=$T
SED=sed

getpath()
{
  PRG=$1

  for p in `echo $PATH | sed -e 's/:/ /g'`
  do
    PP=$p/$PRG
    [ -x $PP ] || continue
    echo $PP
    return 0
  done
  return 1
}

PKILL=`getpath pkill`
[ x"$PKILL" = "x" ] && PKILL="echo -- "

# get nameserver
if test X"${MTA_NAMESERVER}" != "X"
then
  NS=${MTA_NAMESERVER}
else
  NS=`${GREP} '^nameserver' /etc/resolv.conf | head -n1 | sed -e 's/nameserver[	 ]*\([0-9][0-9\.]*\).*$/\1/'`
fi

# for debugging on some *BSD; how about others?
#export MALLOC_OPTIONS="J"

DSN_ERR="A mail from you could not be delivered. See below for details."
DSN_DEL="A mail from you has not yet been successfully delivered.
It will be tried again, so you do not need to resend it!
See below for details."
DSN_BOUNCE="A bounce message could not be delivered.
See below for details."


VERBOSE=${MTA_VERBOSE:-false}
MM=../libsmmap/t-mm
IDBR0=../libibdb/t-idbr-0

ALI=aliases
ALIMAP=${ALI}.db
ACC=access
ACCMAP=${ACC}.db
QMGRCONF=qmgr_conf
QMGRCONFMAP=${QMGRCONF}.db
MT=mt
MTMAP=${MT}
#MTMAP=${MT}.db

#XREF
SMTP_R_QUICK=0x00010000
#XREF
PMEXIT=999

DKTIME=1258487282

SHELL=/bin/sh
test -x /bin/ksh && SHELL=/bin/ksh

I=""
VERBOSE=false
ERRS=0
ERR=false
ERRORPAT="ERROR|FATAL|CRIT|ALERT"
FATALPAT="FATAL|CRIT|ALERT"

SRC=${SD}/../smar/common.sh
if test -x ${SRC}
then
. ${SRC}
fi
if test x"${BASEOUT}" = "x"
then
  echo "$0: $BASEOUT not set: internal/setup error"
  exit 1
fi
EDBROK="${BASEOUT}.edbr"

QOUT=tmp.$$
rm -f ${QOUT}
../qmgr/qmgr -VVV > ${QOUT} 2>/dev/null
if ${GREP} 'QMGR_TEST_NO_AQUPDATE' ${QOUT} >/dev/null
then
  echo "$0: smtps compiled with QMGR_TEST_NO_AQUPDATE: tests cannot work"
  rm -f ${QOUT}
  exit 0
else
  rm -f ${QOUT}
fi

# regex setup
if grep 'define HAVE_REGEX_POSIX' ../include/smconf.h >/dev/null
then
  WB="[[:<:]]"
  WE="[[:>:]]"
else
  WB="\<"
  WE="\>"
fi

# hack...
PGREP=/usr/bin/pgrep
[ -x $PGREP ] || PGREP=true

# ----
chkcto()
{
  PRG=$1
  CTO=$2
  shift 2

  if ../$PRG/$PRG -VVV | ${GREP} "$CTO" >/dev/null
  then
    :
  else
    echo "SKIPPED: $0: $PRG not compiled with $CTO $@"
    exit 0
  fi
}

# ----
hascto()
{
  PRG=$1
  CTO=$2
  shift 2

  ../$PRG/$PRG -VVV | ${GREP} "$CTO" >/dev/null
}

sinkhastls()
{
  [ x"${SINKPRG}" = "x" ] && SINKPRG=${STE}/smtps3
  if ${SINKPRG} -h 2>&1 | fgrep "S: STARTTLS" >/dev/null
  then
    :
  else
    echo "SKIPPED: $0: ${SINKPRG} not compiled with STARTTLS support"
    exit 0
  fi
}

srchastls()
{
  [ x"${SRCPRG}" = "x" ] && SRCPRG=${STE}/smtpc2
  if ${SRCPRG} -h 2>&1 | fgrep "S: STARTTLS" >/dev/null
  then
    :
  else
    echo "SKIPPED: $0: ${SRCPRG} not compiled with STARTTLS support"
    exit 0
  fi
}

smtpctlsdebug()
{
if hascto smtpc MTA_TLS_DEBUG
then
  TLSCDBG="tls { debug=9; }"
else
  TLSCDBG=""
fi
}

# ----
hasipv6()
{
  CTO=MTA_NETINET6

  HAS_MTA_NETINET6=false
  for PRG in qmgr smar smtpc smtps
  do
    ../$PRG/$PRG -VVV | ${GREP} "$CTO" >/dev/null || return 1
  done
  HAS_MTA_NETINET6=true
  return 0
}

# ----
clearlogs()
{
  for i in ${MCPLOGS}
  do
    [ -f ${i} ] && cp /dev/null ${i}
  done
}

# ----
# wait until everything is delivered
alldelivered()
{
  CNT=1
  ST=qs.$$
  rm -f $ST
  while true
  do
    ls ?/S* >$ST 2>/dev/null
    [ -s $ST ] || break
    sleep 1
    CNT=`expr $CNT + 1 `
    if test $CNT -gt 300
    then
      echo "$0: test ${TEST}: waited too long for queue to empty"
      ERR=true
      break
    fi
  done
  rm -f $ST
}

# ----
# create access map
crtacc()
{
	rm -f ${ACC} ${ACCMAP}
	if [ x"$1" != "x" ]
	then
		echo "$1" > ${ACC}
	else
		cp /dev/null ${ACC}
	fi
	${MM} -t'	' -F ${ACCMAP} < ${ACC}
}

# ----
# create qmg_conf map
crtqcnf()
{
	rm -f ${QMGRCONF} ${QMGRCONFMAP}
	if [ x"$1" != "x" ]
	then
		echo "$1" > ${QMGRCONF}
	else
		cp /dev/null ${QMGRCONF}
	fi
	${MM} -t'	' -F ${QMGRCONFMAP} < ${QMGRCONF}
}

# ----
sigpids()
{
  SIG=$1

  if test -s ${PIDS}
  then
  for i in `cat ${PIDS}`
  do
    kill $SIG ${i}
  done
  fi
}

# ----
myexit()
{
  RC=$1

  if test X"${MTA_STOPONERROR}" != X
  then
    echo $0:$RC > stop
  fi
  exit $RC
}


# ----
testend()
{
for arg
do
  case "${arg}" in
   rmb)
	cd ${THISDIR}
	rm -f ${MTMAP} ${ALIMAP} ${ACCMAP}
	cd ../qmgr
	rm -f ${MTMAP} ${ALIMAP} ${ACCMAP}
	;;
  esac
done

test ${ERRS} -eq 0 && exit 0
echo "$0: ${ERRS} error(s)"
exit 1
}

# ----
op_no_ticket()
{
# use SSL_OP_NO_TICKET for all versions before 1.0
if ../smtps/smtps -VV | grep 'OpenSSL: linked=0x0' >/dev/null
then
  TICKET="SSL_OP_NO_TICKET"
  CCONF=c$$.conf
  rm -f ${CCONF}
  cat > ${CCONF} <<EOF
  smtpc {
	tls { cache_size = 8;
		options = { SSL_OP_ALL, ${TICKET} };
	}
  }
EOF
  # is the option actually available?
  if ../smtpc/smtpc -f ${CCONF} -VVVV >/dev/null 2>&1
  then
    :
  else
    TICKET=""
  fi
  rm -f ${CCONF}
else
  TICKET=""
fi
export TICKET
}

# ----
lncert()
{
  C=$1
  [ -s $C ] || return 1
  H=`openssl x509 -noout -hash < $C`.0
  rm -f $H
  ln -s $C $H
  return 0
}

# ----
getcerts()
{
# get certs
for i in smcert.pem smkey.pem CAcert.pem sccert.pem sckey.pem C2cert.pem mtaaltcert.pem mtaaltkey.pem cert4.pem key4.pem
do
  [ -f ${i} -a ! -w ${i} ] && chmod u+w ${i}
  cp ${SD}/${i} .
done

[ -d certs ] || mkdir certs
for i in 0 1 2 3
do
  cac=CA${i}.cert.pem
  [ -f certs/${cac} -a ! -w certs/${cac} ] && chmod u+w  certs/${cac}
  cp ${SD}/${cac} certs
done
(cd certs >/dev/null 2>&1 && lncert CA0.cert.pem)
}

# ----
tlsaliases()
{
# create aliases for tls tests
# Note: these are the union of (almost?) all tests.
${CAT} > ${ALI} <<EOF
abuse:	local:
postmaster:	local:
user:	local:
user1:	local:
user2:	local:
sender:	local:
sender1:	local:
sender2:	local:
sender3:	local:
sender4:	local:
sender5:	local:
sender6:	local:
sender7:	local:
sender8:	local:
sender9:	local:
a2:	<user1@local.dom> <user2@local.dom>
EOF
# create map
rm -f ${ALIMAP}
${MM} < ${ALI}
}

# ----
setuptls0()
{

EM=emailAddress
if ../smtps/smtps -VV | ${GREP} 'compiled against.*=0x009060' >/dev/null
then
  EM=Email
fi

tlsaliases

# create mailertable (save old version?)
${CAT} > ${MT} <<EOF
local.dom	lmtp:
other.dom	[127.1.2.3]
host.local	[127.0.0.1]
EOF

getcerts
}

# ----
setuptls2()
{
MTA_CDQMGRDIR=false
export MTA_CDQMGRDIR
MYV=${VERBOSE}
. ${SD}/mta-setup.sh
VERBOSE=${MYV}
MTA_CDQMGRDIR=true

SNKPORT=8753
SRVPORT=1579
MTA_SNKPORT=8752
MTA_SRVPORT=${SNKPORT}
SNKPORT2=${MTA_SNKPORT}
SRVPORT2=${MTA_SRVPORT}
}

# ----
setuptls1()
{
# create mailertable (save old version?)
${CAT} > ${MT} <<EOF
local.dom	[127.0.0.1]
other.dom	[${MTA_LMTPIPV4D}]
EOF
tlsaliases
getcerts
setuptls2
}

# ----
# common "startup" sequence of t-tls-reqr-* scripts
tlsrintro()
{
chkcto smtps MTA_USE_TLS
chkcto smtpc MTA_USE_TLS

if test X"`uname`" != "XOSF1"
then
while getopts s:V FLAG
do
  case "${FLAG}" in
    s) SKIP="${SKIP} ${OPTARG}";;
    V) VERBOSE=true;;
  esac
done
shift `expr ${OPTIND} - 1`
# linux: killall kills qmgr/smar from first MTA, hence subsequent tests fail
if test X"`uname`" = "XLinux" -a X"${SKIP}" = "X"
then
    SKIP=" 2 3 4 5 6 7 8 9 "
fi
fi

# cd to working directory for alias file!
cd ../qmgr || exit 1
setuptls0
}

# ----
# more setup for tls-reqr tests
# parameters:
# 1: client: qmgr { smtpc { options 
# 2: client: smtpc { options 
# 3: client: access map entry
# 4: server: smtps { options
# 5: server: qmgr { options 
setuptlsr0()
{
TLSR_C_QMSC="$1"
TLSR_C_SMTPC="$2"
TLSR_S_SMTPS="$4"
TLSR_S_QMGR="$5"

# ----------------------------------------
# need two SMTP servers:
# server1: chkmts, uses smx2.conf, started here "by hand"
# server2: qmgr, uses smx.conf, started by smtpcnf.sh
#
# test client [smtpcnf.sh] -> SMTP server 1 (SRVPORT)
# -> QMGR 1 -> SMTP client 1 (SNKPORT)
# -> SMTP server 2 [smtpcnf.sh] (SRVPORT2=SNKPORT)
# -> QMGR 2 [smtpcnf.sh] -> SMTP client 2 [smtpcnf.sh]
# -> lmtp sink [smtpcnf.sh]
# problem: test client in smtpcnf.sh will use SNKPORT (=SRVPORT2)
# hence the port must be passed explicitly via -F ${SRVPORT}
#
# relaying restrictions are imposed in smx.conf

cd ${THISDIR} || exit 1
getcerts
setuptls2
# create mailertable (save old version?)
${CAT} > ${MT} <<EOF
source.dom	lmtp:
local.dom	[127.0.0.1]
other.dom	[${MTA_LMTPIPV4D}]
EOF
# create aliases (source)
${CAT} > ${ALI} <<EOF
abuse:	local:
postmaster:	local:
sender1:	local:
sender2:	local:
sender3:	local:
sender4:	local:
sender5:	local:
sender6:	local:
sender7:	local:
sender8:	local:
sender9:	local:
a2:	<user1@local.dom> <user2@local.dom>
EOF
# create map (source)
rm -f ${ALIMAP}
${MM} < ${ALI}

# ----------------------------------------
# start first MTA (chkmts, it will send the e-mails)
# Note: this MTA keeps running for all the tests! Its configuration
# can't be changed, but the access map can...
${CAT} > ${CONF2} <<EOF
qmgr { Log_Level = 12; wait_for_server=4; wait_for_client=4;
  smtpc { flags = lookup_rcpt_conf;
       ${TLSR_C_QMSC}
      }
}
smar { Log_Level = 12;
}
smtpc { Log_Level = 15; remote_port=${SNKPORT}; wait_for_server=4;
  flags = talk_to_myself; # don't check "talking to myself"
  ${TLSR_C_SMTPC}
}
smtps {
Log_Level = 12;
daemon_address = localhost:${SRVPORT};
wait_for_server=4;
}
EOF

if [ X"$3" != "X" ]
then
${CAT} > ${ACC} <<EOF
$3
EOF
else
${CAT} > ${ACC} <<EOF
smtpc_rcpt_conf:user@local.dom	tls_requirements {cert_issuer="/C=US/ST=California/L=Berkeley/O=Example+20Org/OU=MTA/CN=CA/$EM=MISMATCH@example.org";}
EOF
fi
rm -f ${ACCMAP}
${MM} -t'	' -F ${ACCMAP} < ${ACC}

# version of output files
V=1
# path to programs
P=..
rm -f ${PID2}
trap killittls EXIT 2 15
# go for it
${P}/smar/smar -f ${CONF2} -d9 > a${V}.log 2>&1 &
SMARPID=$!
echo $! >> ${PID2}
${P}/qmgr/qmgr -f ${CONF2} -d9 > q${V}.log 2>&1 &
echo $! >> ${PID2}
${P}/smtpc/smtpc -f ${CONF2} -d9 > c${V}.log 2>&1 &
SMTPC2PID=$!
echo ${SMTPC2PID} >> ${PID2}
${P}/smtps/smtps -f ${CONF2} > s${V}.log 2>&1 &
echo $! >> ${PID2}

# check that the MTA is running
export P
${CHKD}/mta-running.sh ${SRVPORT} || killittls

export MTA_SNKPORT MTA_SRVPORT

cd ../qmgr
${CAT} > ${SMXCNF} <<EOF
qmgr { Log_Level = 12; wait_for_server=4; wait_for_client=4;
 ${TLSR_S_QMGR}
 }
smar { Log_Level = 12; nameserver = ${NS}; }
smtpc { Log_Level = 12; remote_port=${SNKPORT}; wait_for_server=4; }
smtps {
Log_Level = 12;
daemon_address = localhost:${SRVPORT2};
flags = {access};
wait_for_server=4;
${TLSR_S_SMTPS}
}
EOF

${CAT} > ${ACC} <<EOF
EOF
rm -f ${ACCMAP}
${MM} -t'	' -F ${ACCMAP} < ${ACC}
}

# ----
# terminate processes in most TLS tests
killittls()
{
# try to "preserve" any error
[ $? -ne 0 ] && ERRS=`expr ${ERRS} + 1 `
cd ${THISDIR}
if test -s ${PID2}
then
  # stop MTA components
  for i in `${CAT} ${PID2}`
  do
    kill ${i}
  done
fi
# thanks kids for not following the POSIX standard...
if test X"`uname`" = "XLinux"
then
  killall qmgr smar
fi
exit ${ERRS}
}

# ----
# check whether TLS is in a file (usually lmtpdir/?/S*)
chkTLS()
{
  rchkTLS=0
  for B
  do
    if test -s ${B}
    then
      if fgrep 'TLS' ${B} >/dev/null
      then
        :
      else
        ${VERBOSE} && echo "$0: test ${TEST}: TLS not found in ${B}"
        ERR=true
        rchkTLS=1
      fi
    else
      ${VERBOSE} && echo "$0: test ${TEST}: ${B} not found"
      ERR=true
      rchkTLS=1
    fi
  done
  return ${rchkTLS}
}

# ----
# check whether "Received: .* with ESMTPS" is in a file
chkesmtps()
{
  rchkTLS=0
  for B
  do
    if test -s ${B}
    then
      if grep '^Received: by [^ ]* with ESMTPS.$' ${B} >/dev/null
      then
        :
      else
        echo "$0: test ${TEST}: ESMTPS not found in ${B}"
        ERR=true
        rchkTLS=1
      fi
    else
      echo "$0: test ${TEST}: ${B} not found"
      ERR=true
      rchkTLS=1
    fi
  done
  return ${rchkTLS}
}

# ----
# check whether "Received: .* with ESMTPS" is not in a file
chknoesmtps()
{
  rchkTLS=0
  for B
  do
    if test -s ${B}
    then
      if grep '^Received: by [^ ]* with ESMTPS.$' ${B} >/dev/null
      then
        echo "$0: test ${TEST}: ESMTPS found in ${B}"
        ERR=true
        rchkTLS=1
      fi
    else
      echo "$0: test ${TEST}: ${B} not found"
      ERR=true
      rchkTLS=1
    fi
  done
  return ${rchkTLS}
}

# ----
# check whether "verify=" is in a file
chkverify()
{
  V=$1
  shift

  rchkTLS=0
  for B
  do
    if test -s ${B}
    then
      if grep "TLS=.*, verify=$V)" ${B} >/dev/null
      then
        :
      else
        echo "$0: test ${TEST}: verify=$V not found in ${B}"
        ERR=true
        rchkTLS=1
      fi
    else
      echo "$0: test ${TEST}: ${B} not found"
      ERR=true
      rchkTLS=1
    fi
  done
  return ${rchkTLS}
}


# ----
getip2()
{
N=$1

# we don't want localhost
HOSTS=`${SD}/../check2/ips.sh -l`
# start counting at 1 to make up for the removal of localhost
NH=1
for i in ${HOSTS}
do
  NH=`expr $NH + 1 `
done
if test $NH -lt $N
then
  echo "$0: not enough IP addresses: got=$NH, need=$N"
  exit 0
fi
H2=`echo $HOSTS | cut -d' ' -f 1`
}

# ----
# check for something in SCL
chkc()
{
  if ${EGREP} "$@" ${SCL} >/dev/null
  then
    return 0
  else
    $VERBOSE && echo "$0: test {TEST}: missing $@ in ${SCL}"
    ERR=true
    return 1
  fi
}

# ----
# get the value for a $1= in $2 (log entry)
getle()
{
  K=$1
  L=$2
  E=`printf "%s\n" "$L" | sed -e "s/^.* $K=\([^,]*\).*$/\1/"`
  echo $E
}

# ----
# prepare some data for DSN matching
preparedsn()
{
  # argument defaults to B1
  prepdsnf=$1
  [ x"${prepdsnf}" = "x" ] && prepdsnf=${B1}

  [ -s ${prepdsnf} ] || return 1
  # get the MIME delimiter
  DELIM=`${EGREP} '^Content-Type: multipart/report; report-type=delivery-status; boundary=' ${prepdsnf} |\
	${SED} -e 's|^Content-Type: multipart/report; report-type=delivery-status; boundary="\([^"]*\)".*$|\1|'`

  # remove some lines that vary between tests/systems
  ${EGREP} -v '^Date: ' ${prepdsnf} |\
	${SED} -e '/^Received:/,$d' \
	-e 's/
$//' -e 's/Mailer-Daemon@.*/Mailer-Daemon@/' \
	> ${OUT}
}

# ----
# check whether two files are the same
chkdsn()
{
  # argument defaults to B1
  chkf=$1
  [ x"$chkf" = "x" ] && chkf=${B1}

  if ${DIFF} -w ${OK} ${OUT} >/dev/null
  then
    :
  else
    ${VERBOSE} && echo "$0: diff ${OK} ${OUT}"
    ${VERBOSE} && ${DIFF} -w ${OK} ${OUT}
    ERR=true
  fi
}

