#!/bin/sh
# $Id: t-tls-retry-0.sh,v 1.19 2014/02/16 01:00:44 ca Exp $
#
# Copyright (c) 2014 Claus Assmann
#
# By using this file, you agree to the terms and conditions set
# forth in the license/LICENSE.3C file which can be found at the
# top level of this source code distribution.
#
# ----------------------------------------
# avoid TLS after initial failure
# uses smtps2 with TLS
# run smtpcnf.sh with different parameters
# ----------------------------------------
#

SD=`dirname $0`
if test "${SD}" = "."
then
   SD=../chkmts
fi
. ${SD}/common.sh

SINKPRG=${STE}/smtps3
chkcto smtpc MTA_USE_TLS
chkcto smtpc FFR_CERT_PINNING
sinkhastls
if hascto smtpc MTA_TLS_DEBUG
then
  TLSDBG="tls { debug=9; }"
else
  TLSDBG=""
fi
if hascto smtpc FFR_TLS_RETRY_IN_SMTPC
then
  SCTEMPFAIL=465
else
  SCTEMPFAIL=
fi

if test X"`uname`" != "XOSF1"
then
while getopts s:V FLAG
do
  case "${FLAG}" in
    s) SKIP="${SKIP} ${OPTARG}";;
    V) VERBOSE=true;;
  esac
done
shift `expr ${OPTIND} - 1`
fi

# test 6 is broken right now
[ x"${SKIP}" = "x" ] && SKIP=" 6 "

SRCPRG=${STE}/smtpc2
SENDER=sender@local.dom
RCPT=user@local.host
CIPHER="SRP-DSS-AES-128-CBC-SHA"
# HACK!!! ALL | NO_SSLvX | NO_TSLv1[_1]
TLS1_2=0x97000BFF

EDBROK="../qmgr/t-edbr.ok"
cd ../qmgr || exit 1

crtedbrok()
{
rs=$1
# XREF: SMTPC_SE_OP_TLSHS
[ x"${rs}" = "x" ] && rs=469
rm -f ${EDBROK}
${CAT} > ${EDBROK} <<EOF
got transaction
	mail=<${SENDER}>
	rcpts_tot=1
	rcpts_left=1
	rcpts_temp=1
	rcpts_perm=0
	state=469
	aqt_rcpts_ar=0
got recipient
	rcpt=<${RCPT}>
	aqr_rcpt_idx=0
	aqr_da_idx=0
	aqr_status=${rs}
EOF
}

crtcnf()
{
${CAT} > ${SMXCNF} <<EOF
qmgr { Log_Level = 12; wait_for_server=4; wait_for_client=4;
  $1
}
smar { Log_Level = 12; nameserver = ${NS}; }
smtpc { Log_Level = 12; remote_port=${SNKPORT}; wait_for_server=4;
${TLSCDBG}
}
smtps {
Log_Level = 14;
daemon_address = localhost:${SRVPORT};
wait_for_server=4;
flags = {access};
}
EOF
}

rm -f ${ALIMAP} ${MTMAP} ${ACCMAP}
# create aliases
${CAT} > ${ALI} <<EOF
abuse:	local:
postmaster:	local:
user:	local:
sender: local:
EOF
# create map
${MM} < ${ALI}
crtacc
crtqcnf

crtcnf "retry_delay_min = 10s; retry_delay_max = 60s;
  smtpc { flags = {lookup_rcpt_conf,lookup_session_conf,} }"

# ----------------------------------------
# test: send mails to local address; fails due to cipher mismatch,
# next attempt wo STARTTLS
# (may require test whether cipher is available for smtps2?)
TEST=1
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
rm -f ${EDBROK}
crtacc
crtqcnf "smtpc_session_conf:127.	tls_requirements {flags = {once_only} }"
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4 -Yl${CIPHER}" -E 1 -A 1 -X 1 -S${SENDER} -R${RCPT} -s 12
then
  chknoesmtps ${S1}
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi

# ----------------------------------------
# test: send mails to local address; fails due to cipher mismatch
# next attempt fails due to tls_requirements
TEST=2
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
crtacc "smtpc_rcpt_conf:${RCPT}	tls_requirements {flags = {verified}; cipher_bits_min=256; }"
crtqcnf "smtpc_session_conf:127.	tls_requirements {flags = {once_only} }"
crtedbrok 421
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4 -Yl${CIPHER}" -E 0 -A 1 -X 1 -S${SENDER} -R${RCPT} -s 12 -O ${EDBROK}
then
  rm -f ${EDBROK}
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi

# ----------------------------------------
# test: send mails to local address; fails due to cipher mismatch
# next attempt fails due to tls_requirements requiring always to try TLS
TEST=3
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
crtacc
crtqcnf "smtpc_session_conf:127.	tls_requirements {flags = {always_try} }"

crtedbrok ${SCTEMPFAIL}
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4 -Yl${CIPHER}" -E 0 -A 1 -X 0 -S${SENDER} -R${RCPT} -s 12 -O ${EDBROK}
then
  rm -f ${EDBROK}
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi

# ----------------------------------------
# test: send mails to local address;
# server speaks only TLSv1_2
# XXX this succeeds on first try!!!
# how to make it initially fail?
# if we force it to SSLv3 then smtpc won't try again with something different
# as we don't override options...
TEST=4
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
crtacc
crtqcnf
crtcnf "retry_delay_min = 2s; retry_delay_max = 4s;"

rm -f ${EDBROK}
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4 -Yo${TLS1_2}" -E 1 -A 1 -X 1 -S${SENDER} -R${RCPT} -s 24
then
  rm -f ${EDBROK}
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi

# ----------------------------------------
# test: send mails to local address; fails due to cipher mismatch,
# retry several combinations (how to check?)
# until delivery final succeeds wo STARTTLS
TEST=5
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
rm -f ${EDBROK}
crtacc
crtqcnf
crtcnf "retry_delay_min = 2s; retry_delay_max = 4s;"
if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d7 -Yl${CIPHER}" -E 1 -A 1 -X 1 -S${SENDER} -R${RCPT} -s 9
then
  chknoesmtps ${S1}
else
  ERR=true
fi
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi


# ----------------------------------------
# test: send mails to local address; fails initially due to cipher mismatch,
# but a new server is started and the occ entry times out before the next try.
# this breaks if smtpc is retrying the connections itself...
TEST=6
if echo "${SKIP} X" | ${GREP} -v " ${TEST} " > /dev/null
then
ERR=false
crtcnf "retry_delay_min = 20s; retry_delay_max = 60s;"
crtacc
crtqcnf "octo:	5
smtpc_session_conf:127.		tls_requirements {flags = { once_only}}"
crtedbrok

if ${SHELL} ${SD}/smtpcnf.sh -r -Y "-@S -d4 -Yl${CIPHER}" -E 0 -A 1 -X 0 -S${SENDER} -R${RCPT} -z -O ${EDBROK}
then

  # stop smtps3, change cipherlist
  SRVPID=`${TAIL} -n 1 ${PIDS}`
  if test X"${SRVPID}" != "X"
  then
    kill -TERM ${SRVPID}
  else
    echo "$0: SRVPID not found?"
  fi

# start SMTP sink
${SINKPRG} -C ./${SMTPDIR} -@S -d4 -l . -b ${H}:${SNKPORT} >> ${SRVL} 2>&1 &
status=$?
SRVPID=$!
if test "x${status}" != "x0"
then
  echo "$0: FAIL: cannot start server ${status}"
  exit ${status}
fi
sleep 1

  # wait for qmgr to retry
  sleep 18
  chkesmtps ${S1}
  rm -f ${EDBROK}

else
  ERR=true
fi
${SD}/stopmts.sh
test X"${SRVPID}" != "X" && kill ${SRVPID}
if ${ERR}
then
   echo "$0: test ${TEST} failed"
   ERRS=`expr ${ERRS} + 1 `
   test X"${MTA_STOPONERROR}" != X && exit 1
fi
fi


# ----------------------------------------
rm -f ${MTMAP} ${ALIMAP} ${QMGRCONFMAP} ${QMGRCONF} ${EDBROK}

testend
