/*
 * Copyright (c) 2011 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-rcpt-perf-0.c,v 1.3 2011/03/03 00:59:44 ca Exp $")

#include <stdio.h>

#include "sm/assert.h"
#include "sm/string.h"
#include "sm/sysexits.h"
#include "sm/test.h"
#include "sm/actdb-int.h"
#include "adb.h"

uint qm_debug = 0;

static void
test(uint maxn, uint nodes, bool check)
{
	int a;
	uint u;
	sm_ret_T ret;
	sessta_id_T ta_id;
/*	rcpt_idx_T rcpt_idx; */
	aq_ctx_P aq_ctx;
	aq_ta_P aq_ta;
	aq_rcpt_P aq_rcpt;

	ret = aq_open(NULL /*qmgr_ctx*/, &aq_ctx, maxn, AQ_OPEN_FL_NOHT);
	SM_TEST_E(ret == SM_SUCCESS);

	for (u = 0, a = 0; u < nodes; u++) {
		sm_snprintf(ta_id, sizeof(ta_id), SMTPS_STID_FORMAT,
			(ulonglong_T) a, 0);
		ret = aq_ta_add_new(aq_ctx, &aq_ta, 0, 1, THR_LOCK_UNLOCK);
		if (maxn == 0 || u < maxn) {
			if (sm_is_err(ret))
				fprintf(stderr, "add: u=%u, a=%d, ta_id=%s, ret=%#x\n", u, a, ta_id, ret);
			SM_TEST_E(!sm_is_err(ret));
		}
		else
			SM_TEST_E(sm_is_err(ret));
		SESSTA_COPY(aq_ta->aqt_ss_ta_id, ta_id);
		ret = aq_tht_add(aq_ctx, aq_ta);
		SM_TEST_E(!sm_is_err(ret));
		ret = aq_rcpt_add_new(aq_ctx, aq_ta, &aq_rcpt, 0, THR_LOCK_UNLOCK);
		SM_TEST_E(!sm_is_err(ret));

		SESSTA_COPY(aq_rcpt->aqr_ss_ta_id, ta_id);
		aq_rcpt->aqr_idx = a;
		ret = aq_rht_add(aq_ctx, aq_rcpt);
		SM_TEST_E(!sm_is_err(ret));

		++a;
	}

	for (u = 0, a = 0; u < nodes; u++) {
		sm_snprintf(ta_id, sizeof(ta_id), SMTPS_STID_FORMAT,
			(ulonglong_T) a, 0);
		aq_rcpt = aq_rht_find_s(aq_ctx, ta_id, SMTP_RCPTIDX_MAX);
		SM_TEST_E(NULL != aq_rcpt);
		aq_rcpt = aq_rht_find_r(aq_ctx, ta_id, a);
		SM_TEST_E(NULL != aq_rcpt);

/*
		ret = aq_ta_find(aq_ctx, ta_id, true, &aq_ta);
		if (sm_is_err(ret))
			fprintf(stderr, "find: u=%u, a=%d, ta_id=%s, ret=%#x\n", u, a, ta_id, ret);
		SM_TEST_E(!sm_is_err(ret));
*/

		++a;
	}

  error:
	aq_close(aq_ctx);
	return;
}

int
main(int argc, char *argv[])
{
	int c;
	uint maxn, nodes;
	bool check;

	nodes = 10000;
	maxn = nodes * 2;
	check = true;

	while ((c = getopt(argc, argv, "ch:m:s:")) != -1) {
		switch (c) {
		  case 'c':
			check = false;
			break;
		  case 'h':
			c = strlen(optarg);
#if 0
			if (c >=  - 1) {
				uint h;

				h = aq_hash(optarg);
				fprintf(stderr, "h=%u\n", h);
				return 0;
			}
			else
				fprintf(stderr, "len(%s)=%d\n", optarg, c);
			return -1;
#endif
			break;
		  case 'm':
			maxn = strtoul(optarg, NULL, 0);
			break;
		  case 's':
			nodes = strtoul(optarg, NULL, 0);
			break;
		  default:
			return EX_USAGE;
		}
	}

	sm_test_begin(argc, argv, "test aq perf 0");
	test(maxn, nodes, check);
	return sm_test_end();
}
