/*
 * Copyright (c) 2006 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-conf-7.c,v 1.8 2013/10/23 05:10:23 ca Exp $")

#if SM_LIBCONF_ALONE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sm-conf.h"
#else /* SM_LIBCONF_ALONE */
#include "sm/io.h"
#include "sm/string.h"
#include "sm/net.h"
#include "sm/units.h"
#include "sm/sm-conf.h"
#define SM_CONF_BYTE_DEF 1
#define SM_CONF_KBYTE_DEF 1
#include "sm/sm-conf-byte.h"
#include "sm/sm-conf-prt.h"
#include <stdio.h>
#endif /* SM_LIBCONF_ALONE */

#ifndef SM_DNS_TSKS_MAX
# define SM_DNS_TSKS_MAX     4u
#endif


int
chk(sm_conf_T *handle, void	*callback_data, sm_conf_definition_T const	*definition, void const	*value_data)
{
	sm_ret_T ret;
	const char **ips;
	const char *ip;
	char *end;
	sm_ipaddr_T ipaddr;

	if (NULL == value_data)
		return SM_CONF_ERR_INVALID;
	ips = (const char **)value_data;
	for (ip = *ips; ip != NULL; ip = *(++ips)) {
		ret = sm_inet_a2ip(ip, &end, &ipaddr);
/*
		sm_io_fprintf(smioerr, "ip=%s, ret=%r, ip=%I\n", ip, ret, &ipaddr);
*/
		if (sm_is_err(ret))
			return SM_CONF_ERR_SYNTAX;
			/* return ret; */
	}

	return 0;
}


/*
**  test argv
*/

typedef struct ss_cnf_S		ss_cnf_T, *ss_cnf_P;
struct ss_cnf_S
{
	char			*ss_cnf_nameservers[SM_DNS_TSKS_MAX];
};

sm_conf_definition_T ss_defs[] = {
{ SM_CONF_DEF_MAGIC, "nameservers",		sm_conf_type_argv,
	offsetof(ss_cnf_T, ss_cnf_nameservers),	SM_DNS_TSKS_MAX,
	NULL,	0,	NULL,	chk,	NULL,
	"List of nameservers (IPv4 addresses)"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },
CONF_SENTINEL
};

static ss_cnf_T s;

static int
process(char const *name, FILE *fp)
{
	sm_conf_T		*smc;
	int			err;
	char buf[SM_CONF_ERROR_BUFFER_SIZE];
	char const *e = NULL;

	if (((smc = sm_conf_new(name ? name : "*stdin*"))) == NULL) {
		fprintf(stderr, "error -- sm_conf_new() returns NULL!\n");
		return 1;
	}
	if ((err = sm_conf_read_FILE(smc, name, fp)) != 0) {
		fprintf(stderr, "%s: %s\n",
			name ? name : "*stdin*",
			sm_conf_strerror(err, buf, sizeof buf));
		while ((e = sm_conf_syntax_error(smc, e)) != NULL)
			fprintf(stderr, "%s\n", e);
		sm_conf_destroy(smc);
		return 2;
	}

	if ((err = sm_conf_scan(smc, ss_defs, 0, &s)) != 0) {

		fprintf(stderr, "(while scanning) %s: %s\n",
			name ? name : "*stdin*",
			sm_conf_strerror(err, buf, sizeof buf));
		while ((e = sm_conf_syntax_error(smc, e)) != NULL)
			fprintf(stderr, "%s\n", e);
		sm_conf_destroy(smc);
		return 3;
	}

	(void) sm_conf_prt_conf(ss_defs, &s, smioerr);

#if 0
	sm_conf_destroy(smc);
#endif

	return 0;
}

int
main(int ac, char **av)
{
	int ai, c, ret, done;
	char *ssname;

	done = 0;
	ssname = NULL;
	while ((c = getopt(ac, av, "f:N:")) != -1) {
		switch (c) {
		  case 'f':
			ret = process(optarg, NULL);
			if (ret != 0)
				return ret;
			done = 1;
			break;
		  case 'N':
			ssname = strdup(optarg);
			if (ssname == NULL)
				return ENOMEM;
			break;
		}
	}
	ac -= optind;
	av += optind;

	if (ac == 0)
		return process("*stdin*", stdin);

	for (ai = 0; ai < ac; ai++) {
		int ret = process(av[ai], NULL);
		if (ret != 0)
			return ret;
	}
	return 0;
}
