/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-expr-0.c,v 1.10 2012/11/05 15:24:57 ca Exp $")

#include "sm/string.h"
#include "sm/net.h"
#include "sm/io.h"
#include "sm/sm-conf.h"
#include "sm/sm-conf-prt.h"
#define SM_CONF_LOG_DEF 1
#include "sm/logcnfdef.h"
#include <stdio.h>
#include "sm/sm-conf-expr.h"
#include "sm-expr.h"

/*
**  test program for expr
*/

typedef struct
{
	cexpr_P			s_expr;
} structure;

#define T_ID_MAIL	1
#define T_ID_RCPT	2
#define T_ID_NRCPTS	3

sm_id_entry_T idtab[] = {
	  { ID_STR("mail"),		vt_str,  0, T_ID_MAIL, 0 }
	, { ID_STR("nrcpts"),	vt_int,  0, T_ID_NRCPTS, 0 }
	, { ID_STR("one"),		vt_int,  1, 0, IDEF_CONST_INT }
	, { ID_STR("rcpt"),		vt_str,  0, T_ID_RCPT, 0 }
	, { ID_STR("ten"),		vt_int, 10, 0, IDEF_CONST_INT }
	/* for testing lengths */
	, { ID_STR("ten0"),		vt_int, 100, 0, IDEF_CONST_INT }
	, { ID_STR("three"),	vt_int,  3, 0, IDEF_CONST_INT }
	, { ID_STR("twelve"),	vt_int, 12, 0, IDEF_CONST_INT }
	, { ID_STR("two"),		vt_int,  2, 0, IDEF_CONST_INT }

};

sm_conf_definition_T
definitions[] =
{
	{ SM_CONF_DEF_MAGIC, "expr", sm_conf_type_expr,
		offsetof(structure, s_expr), sizeof(cexpr_P),
		NULL, 	0,	NULL, NULL, NULL,
		"expression"
		SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0)
		, idtab, SM_ARRAY_SIZE(idtab)},

CONF_SENTINEL
};

static structure s;

static int
idgetval(
		sm_conf_T		*handle,
		void			*callback_data,
		int				 idx,
		sm_conf_definition_T const	*definition,
		val_type_T		*value_type,
		void 			*value_data,
		uint 			*value_length)
{
	printf("idgetval: idx=%d\n", idx);
	if (idx == T_ID_MAIL) {
#define DUMMY_MAIL	"<sender@example.com>"
		*(char **)value_data = DUMMY_MAIL;
		*value_length = sizeof(DUMMY_MAIL) - 1;
		*value_type = vt_str;
		return 0;
	}

	if (idx == T_ID_RCPT) {
#define DUMMY_RCPT	"<rcpt@example.net>"
		*(char **)value_data = DUMMY_RCPT;
		*value_length = sizeof(DUMMY_RCPT) - 1;
		*value_type = vt_str;
		return 0;
	}

	if (idx == T_ID_NRCPTS) {
		*(int *)value_data = 3;
		*value_type = vt_int;
		return 0;
	}

	return -1;
}

static sm_eval_ctx_T eval_ctx = { idtab, SM_ARRAY_SIZE(idtab), idgetval, NULL };

static void
print_structure(structure *st)
{
	sm_ret_T ret;

	if (NULL == st->s_expr) {
		printf("expr: NULL\n");
		return;
	}
	ret = eval_exp(st->s_expr, st->s_expr->cex_rooti, &eval_ctx);
	printf("expr: %d\n", CXN(st->s_expr, st->s_expr->cex_rooti).cen_eval.cv_int);
	printf("eval: %d\n", ret);
}

static void
eval_it(structure *st, long l)
{
	long i;
	time_t startt, endt;
	sm_ret_T ret;

	if (NULL == st->s_expr)
		return;
	startt = time((time_t *)0);
	for (i = 0; i < l; i++) {
		ret = eval_exp(st->s_expr, st->s_expr->cex_rooti, &eval_ctx);
	}
	endt = time((time_t *)0);
	printf("time=%ld\n", (long) (endt - startt));
}

static int
process(char const *confname, FILE *fp, bool show, bool prt)
{
	sm_conf_T *smc;
	int err;
	char buf[SM_CONF_ERROR_BUFFER_SIZE];
	char const *e = NULL;

	if (((smc = sm_conf_new(confname ? confname : "*stdin*"))) == NULL) {
		fprintf(stderr, "error -- sm_conf_new() returns NULL!\n");
		return 1;
	}
	if ((err = sm_conf_read_FILE(smc, confname, fp)) != 0) {
		fprintf(stderr, "%s: %s\n",
			confname ? confname : "*stdin*",
			sm_conf_strerror(err, buf, sizeof buf));

		while ((e = sm_conf_syntax_error(smc, e)) != NULL)
			fprintf(stderr, "%s\n", e);

		sm_conf_destroy(smc);
		return 2;
	}

	err = sm_conf_scan(smc, definitions,
		SM_CONF_FLAG_ALLOW_ANY|SM_CONF_FLAG_KEEP_DEFAULT, &s);
	if (err != 0) {
		fprintf(stderr, "(while scanning) %s: %s\n",
			confname ? confname : "*stdin*",
			sm_conf_strerror(err, buf, sizeof buf));

		while ((e = sm_conf_syntax_error(smc, e)) != NULL)
			fprintf(stderr, "%s\n", e);

		sm_conf_destroy(smc);
		return 3;
	}

	if (show)
		print_structure(&s);

	if (prt) {
		sm_str_P text;

		text = sm_str_new(NULL, 256, 8192);
		sm_io_fprintf(smioout, "prt_conf1:\n");
		(void) sm_conf_prt_cnfs(definitions, &s, 0, smioout, 0, NULL, text, 0);
		sm_io_flush(smioout);
	}

#if 0
	sm_conf_destroy(smc);
#endif

	return 0;
}

int
main(int ac, char **av)
{
	int ai, c, ret;
	long loops;
	bool prt, done;
	char *option;

	done = false;
	prt = false;
	option = NULL;
	loops = 0;
	while ((c = getopt(ac, av, "df:l:s")) != -1) {
		switch (c) {
		  case 'd':
			sm_conf_prt_dflt(definitions, SMC_FLD_FLAGS, smioout);
			sm_io_flush(smioout);
			return 0;
		  case 'f':
			ret = process(optarg, NULL, 0, prt);
			if (ret != 0)
				return ret;
			done = true;
			break;
		  case 'l':
			loops = atol(optarg);
			break;
		  case 's':
			prt = true;
			break;
		}
	}
	ac -= optind;
	av += optind;
	if (done) {
		print_structure(&s);
		if (loops > 0)
			eval_it(&s, loops);
		if (prt) {
			sm_io_fprintf(smioout, "prt_conf:\n");
			(void) sm_conf_prt_cnfs(definitions, &s, 0,
					smioout, 0, NULL, NULL, 0);
			sm_io_flush(smioout);
		}
		return 0;
	}

	if (ac == 0) {
		ret = process("*stdin*", stdin, 1, prt);
		if (ret == 0 && loops > 0)
			eval_it(&s, loops);
		return ret;
	}

	for (ai = 0; ai < ac; ai++) {

		ret = process(av[ai], NULL, 1, prt);
		if (ret != 0)
			return ret;
		if (loops > 0)
			eval_it(&s, loops);
	}
	return 0;
}
