/*
 * Copyright (c) 2002, 2004, 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: netsockconn.c,v 1.11 2013/09/18 03:15:50 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/fcntl.h"
#include "sm/net.h"

/*
**  NET_CLIENT_CONNECT -- Create a socket and connect it to a server.
**
**	Parameters:
**		ip -- IP to connect to
**		port -- port on host to connect to
**		pfd -- (pointer to) file descriptor (output)
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
net_client_connect(const char *ip, int port, int *pfd)
{
	int sockfd, err;
	sm_ret_T ret;
	sm_sockaddr_T servaddr;
	sockaddr_len_T len;

	SM_REQUIRE(ip != NULL);
	SM_REQUIRE(port > 0);
	SM_REQUIRE(pfd != NULL);

	*pfd = INVALID_SOCKET;

	ret = sm_inet_a2sockaddr(ip, NULL, &servaddr);
	if (sm_is_error(ret))
		return ret;
#if MTA_NETINET6
	if (AF_INET6 == servaddr.sa.sa_family) {
		len = sizeof(servaddr.sin6);
		servaddr.sin6.sin6_port = htons(port);
	}
	else
#endif
	if (AF_INET == servaddr.sa.sa_family) {
		len = sizeof(servaddr.sin);
		servaddr.sin.sin_port = htons(port);
	}
	else
		return sm_error_perm(SM_EM_NET, EPFNOSUPPORT);

	if ((sockfd = socket(servaddr.sa.sa_family, SOCK_STREAM, 0)) == -1)
		return sm_error_perm(SM_EM_NET, errno);

	if ((connect(sockfd, (struct sockaddr *) &servaddr.sa, len)) == -1) {
		err = sm_error_perm(SM_EM_NET, errno);
		(void) close(sockfd);
		return err;
	}
	*pfd = sockfd;
	return SM_SUCCESS;
}
