/*
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: rcbget3uintn.c,v 1.2 2011/01/30 21:49:06 ca Exp $")
#include "sm/limits.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/magic.h"
#include "sm/memops.h"
#include "sm/rpool.h"
#include "sm/rcb.h"
#include "sm/reccom.h"

/*
**  SM_RCB_GET3UINTn -- Read two ints and one buf from a sm_rcb_P.
**
**	Parameters:
**		rcb -- sm_rcb_P object to read int from.
**		pv1 -- pointer to value to return.
**		pv2 -- pointer to value to return.
**		buf -- buf to return.
**		len -- length of buf
**
**	Returns:
**		usual sm_error code; SM_E_OVFLW_NS
*/

sm_ret_T
sm_rcb_get3uintn(sm_rcb_P rcb, uint32_t *pv1, uint32_t *pv2, void *buf, uint len)
{
	SM_IS_RCB(rcb);
#if SM_RCB_CHECK
	SM_REQUIRE(rcb->sm_rcb_state == SM_RCB_DEC);
#endif
#if PARANOID
	SM_REQUIRE(pv1 != NULL);
	SM_REQUIRE(pv2 != NULL);
	SM_REQUIRE(buf != NULL);
	SM_REQUIRE(rcb->sm_rcb_base != NULL);
#endif
	if (rcb->sm_rcb_len < rcb->sm_rcb_rw + sizeof(uint32_t) * 2 + len)
		return sm_error_perm(SM_EM_RECCOM, SM_E_OVFLW_NS);
	sm_buf2uint32(rcb->sm_rcb_base + rcb->sm_rcb_rw, pv1);
	rcb->sm_rcb_rw += sizeof(uint32_t);
	sm_buf2uint32(rcb->sm_rcb_base + rcb->sm_rcb_rw, pv2);
	rcb->sm_rcb_rw += sizeof(uint32_t);
	sm_memcpy(buf, rcb->sm_rcb_base + rcb->sm_rcb_rw, len);
	rcb->sm_rcb_rw += len;
	return SM_SUCCESS;
}
