#!/bin/sh
# $Id: smtp.sh,v 1.10 2013/12/08 02:22:15 ca Exp $
#
# simple test program running SMTP server and client to send and receive
# 1000 messages.
#
# host on which to listen/to which to send
H=localhost
# port number to use
P=${MTA_SRVPORT:-1235}
ERRS=0
SRVL="smtps.log"
CLTL="smtpc.log"
rm -f $SRVL $CLTL
#

for SRV in smtps smtps2
do
# start SMTP server
./$SRV -l . -b $H:$P > $SRVL 2>&1 &
status=$?
SRVPID=$!
if [ "x$status" != "x0" ]
then
  echo "$0: FAIL: cannot start server $status"
  exit $status
fi
sleep 3
if grep "ERROR" $SRVL > /dev/null
then
  echo "$0: found ERROR in $SRVL"
  ERRS=`expr $ERRS + 1 `
  test X"${MTA_STOPONERROR}" != X && exit 1
fi

# send SMTP server a signal to dump its state
if kill -USR1 $SRVPID
then
  :
else
  echo "$0: kill -USR1 failed: $?"
  exit 1
fi
sleep 2

# check state of SMTP server
if grep "^Address *$H:$P" $SRVL > /dev/null
then
  :
else
  echo "$0: failed to listen on $H:$P"
  ERRS=`expr $ERRS + 1 `
fi
if grep "^Thread.* 8/" $SRVL > /dev/null
then
  :
else
  echo "$0: failed to start threads"
  ERRS=`expr $ERRS + 1 `
fi
if grep "^Waiting threads *8" $SRVL > /dev/null
then
  :
else
  echo "$0: wrong number of waiting threads"
  ERRS=`expr $ERRS + 1 `
fi
if grep "^Busy threads *0" $SRVL > /dev/null
then
  :
else
  echo "$0: wrong number of busy threads"
  ERRS=`expr $ERRS + 1 `
fi
if grep "^Sessions served *0" $SRVL > /dev/null
then
  :
else
  echo "$0: wrong number of Sessions"
  ERRS=`expr $ERRS + 1 `
fi

# send 1000 messages: 100 connections, 10 threads
if ./smtpc -r $H:$P -c 100 -t 10 > $CLTL 2>&1
then
  :
else
  cat $CLTL
  ERRS=`expr $ERRS + 1 `
fi

# different recipient domains
if ./smtpc2 -d 5 -Yr1.meta1.org -Yr2.meta1.org -r $H:$P -s 4 -t 4 > $CLTL 2>&1
then
  for r in r1 r2
  do
    if fgrep $r.meta1.org $CLTL > /dev/null
    then
      :
    else
      echo "$0: missing $r as recpient domain."
      ERRS=`expr $ERRS + 1 `
      test X"${MTA_STOPONERROR}" != X && exit 1
    fi
  done
else
  cat $CLTL
  ERRS=`expr $ERRS + 1 `
fi

# different sender domains
if ./smtpc2 -d 5 -Fs1.meta1.org -Fs2.meta1.org -r $H:$P -s 4 -t 4 > $CLTL 2>&1
then
  for s in s1 s2
  do
    if fgrep $s.meta1.org $CLTL > /dev/null
    then
      :
    else
      echo "$0: missing $s as sender domain."
      ERRS=`expr $ERRS + 1 `
      test X"${MTA_STOPONERROR}" != X && exit 1
    fi
  done
else
  cat $CLTL
  ERRS=`expr $ERRS + 1 `
fi

# ----------------
kill -TERM $SRVPID

# wait a moment to release socket
test $SRV = smtps && sleep 5

done

[ $ERRS -eq 0 ] && exit 0
echo "$0: $ERRS error(s)"
exit 1
