/*
 * client SMTP program, use sm io.
 *
 * based on proxy.c
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: smtpc2.c,v 1.162 2014/03/04 05:38:43 ca Exp $")
#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/str.h"
#include "sm/test.h"
#include "sm/io.h"
#include "sm/ctype.h"
#include "sm/net.h"
#include "sm/socket.h"
#include "sm/sysexits.h"
#include "sm/misc.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#if HAVE_MATH_H
#include <math.h>
#define SM_MATH_STATS 1
#else
#define SM_MATH_STATS 0
#endif

#include "error.h"
static int errfd = STDERR_FILENO;

#if MTA_USE_TLS
#include "sm/tls.h"
#include "sm/tlscnf.h"
#include "sm/tlsbio.h"
#include "sm/smtpdef.h"
#include "sm/confsetpath.h"
tls_ctx_P tls_ctx;
tls_cnf_T tls_cnf;
bool try_starttls = false;
#endif

/* SM_ASSERT doesn't trigger? */
#define MY_ASSERT(c)	do { if (!(c)) abort(); } while (0)

#include "st.h"
#include "common.h"	/* HACK for debugging, otherwise structs are hidden */

extern sm_stream_T SmStThrIO;

#define SM_IOBUFSIZE (1024*1024)
#define SM_LINE_LEN (4*1024)
static	uchar databuf[SM_IOBUFSIZE];

#ifndef INADDR_NONE
# define INADDR_NONE 0xffffffff
#endif

#define MAXTC	65536L		/* max. number of total connections */
#define SC_MAXADDRS	4096	/* max. number of addresses */
#define REQUEST_TIMEOUT 5
#define SEC2USEC(s) ((s)*1000000LL)

static char *prog;			/* Program name   */
static sm_sockaddr_T rmt_addr;	/* Remote address */
static unsigned int ta_per_thrd;
static unsigned int sessions;
static unsigned int sess_cnt = 0;
static int debug = 0;
static int stoponerror = 0;
static int contiftaerror = 0;
static unsigned long count = 0;
static int busy = 0;
static int concurrent = 0;
static int waitdata = 0;
static int sequence = -1;
static int seqfirst = 0;
static int postfix = 0;
static int rset = 1;
static int idisplay = 0;	/* interactive display */
static int show_rate = 0;
static bool dontread_QUIT_reply = false;
static unsigned long datalen = 0;
static char *cmdfile = NULL;
#define MAXDF	1024
static char *datafiles[MAXDF];
static bool use_df_perc = false;
static unsigned int df_total = 0;
static unsigned short df_perc[MAXDF];
static unsigned long df_used[MAXDF];
static unsigned int ndatafiles = 0;
static unsigned int ndatafile = 0;
static char *datafile = NULL;
static bool alteventsys = false;

#define MAX_CLT_IPS	1024
static ipv4_T clt_ip;
static ipv4_T clt_ips[MAX_CLT_IPS];
static unsigned int nclt_ips = 0;
static unsigned int nclt_ip = 0;
static bool use_bind = false;

#define MAXHI	1024
static char *hostinfos[MAXHI];
static char *hostinfo;
static unsigned int nhostinfos = 0;
static unsigned int nhostinfo = 0;
static bool use_hi_perc = false;
static unsigned int hi_total = 0;
static unsigned short hi_perc[MAXHI];
static unsigned long hi_used[MAXHI];

static int flushdata = 0;
static unsigned int rcpts = 0;
static unsigned long ta_total = 0;
static unsigned long total_rate = 0;
static bool total_rate_set = false;
static unsigned long ta_cnt = 0;
static unsigned long startup = 0;
static unsigned long startup_act = 0;
static unsigned long teardown = 0;
/* static unsigned long teardown_act = 0; not yet used */
static const char *from[SC_MAXADDRS], *rcpt[SC_MAXADDRS];
static const char *fromdom[SC_MAXADDRS], *rcptdom[SC_MAXADDRS];
static const char *orcpt = NULL;	/* only one for now */
static unsigned int fromaddrs = 0;
static unsigned int rcptaddrs = 0;
static unsigned int fromdoms = 0;
static unsigned int rcptdoms = 0;
static unsigned int request_timeout = REQUEST_TIMEOUT;
static int showerror = 0;
static bool usesize = false;
static char *mailarg = NULL;
static char *ehloarg = "me.local";
static time_t firstNmail = 0;
static time_t firstmail = 0;
static time_t stoptime = 0;

static bool try_drr = false;
static bool use_drr = false;
static bool try_rsad = false;
static bool use_rsad = false;
static bool use_pipelining = false;
static bool delayonerror = false;
static bool rpct_per_ta = false;
static bool uniqueaddrs = false;
static unsigned long addrcnt = 0;
static int random_drop = 0;
#define RANDOM_DROP do { \
	if (random_drop > 0 && random() < random_drop) {	\
		if (debug)	\
			sc_dprintf(tid, "random_drop\n");\
		goto fail;	\
	}	\
	} while (0)

/* # of recipient addresses sent, # of replies received, # of "OK"s */
/* shouldn't most of these be per TA??? (at least per session) */
static sm_str_P reply_str = NULL;

/* total count of MAIL/RCPT commands */
static unsigned long mail_cnt = 0;
static unsigned long rcpt_cnt = 0;

/* 3 digits reply code plus ' ' or '-' */
#define SMTP_REPLY_OFFSET	4

#if SC_USE_DATE
#define MAIL_HEADER "From: me@example.com\r\nTo: you@example.com\r\nSubject: test\r\nDate: today\r\n\r\n"
#else
#define MAIL_HEADER "From: me@example.com\r\nTo: you@example.com\r\nSubject: test\r\n\r\n"
#endif

/* server capabilities */
#define SCSE_CAP_NONE		0x0000	/* guess	*/
#define SCSE_CAP_ESMTP		0x0001	/* ESMTP	*/
#define SCSE_CAP_PIPELINING	0x0010	/* PIPELINING	*/
#define SCSE_CAP_8BITMIME	0x0020	/* 8BITMIME	*/
#define SCSE_CAP_SIZE		0x0040	/* SIZE		*/
#define SCSE_CAP_ENHSTAT	0x0080	/* ENHANCEDSTATUSCODES	*/
#define SCSE_CAP_AUTH		0x0100	/* AUTH		*/
#define SCSE_CAP_STARTTLS	0x0200	/* STARTTLS	*/
#define SCSE_CAP_DRR		0x0400	/* DRR	*/
#define SCSE_CAP_RSAD		0x0800	/* RSAD	*/

#define SCSE_SET_CAP(caps, cap)	(caps) |= (cap)
#define SCSE_CLR_CAP(caps, cap)	(caps) &= ~(cap)
#define SCSE_IS_CAP(caps, cap)	(((caps) & (cap)) != 0)


#define SCE_ST_ABORT_SE	0x0100
#define SCE_ST_ABORT_TA	0x0200

#define SCE_CODE(code)	((unsigned int)((code) & 0x7f00))
#define SCE_VAL(code)	((unsigned int)((code) & 0x00ff))
#define SCE_IS_ABORT_SE(code, counter)	\
	((SCE_CODE(code) == SCE_ST_ABORT_SE) && \
	 (SCE_VAL(code) == 0 || SCE_VAL(code) == (counter)))
#define SCE_IS_ABORT_TA(code, counter)	\
	((SCE_CODE(code) == SCE_ST_ABORT_TA) && \
	 (SCE_VAL(code) == 0 || SCE_VAL(code) == (counter)))

#define STAGE_QUIT	(-2)
#define STAGE_UNKNOWN	(-1)
#define STAGE_CONNECT	0
#define STAGE_HELO	1
#define STAGE_TLS	2
#define STAGE_MAIL	3
#define STAGE_RCPT	4
#define STAGE_DATA	5
#define STAGE_DOT	6
/* #define STAGE_LDAT	7 */
#define STAGE_NEXT	7	/* next command after final dot, i.e., next TA */
#define STAGES		8

#define SC_PIPELINE_OK(stage) (((STAGE_MAIL == (stage)) || (STAGE_RCPT == (stage))) && use_pipelining)

static int codes[STAGES];
static int waits[STAGES];

static void *handle_request(void *arg);
static void print_sys_error(const char *msg);

/*
**  timing statistics
*/

#define ST_SE_CONNECT	0
#define ST_SE_EHLO	1
#define ST_SE_QUIT	3
#define ST_SE_N		4

#define ST_TA_RSET	0
#define ST_TA_MAIL	1
#define ST_TA_RCPT	2	/* what about multiple recipients??? */
#define ST_TA_DATA	3
#define ST_TA_DOT	4
#define ST_TA_N		5

typedef struct sce_stt_S		sce_stt_T, *sce_stt_P;
struct sce_stt_S
{
	st_utime_t	 scet_wr[ST_TA_N];
	st_utime_t	 scet_rd[ST_TA_N];
	st_utime_t	 scet_ta;	/* for one transaction */
};

/* transaction context */
typedef struct sce_ta_S		sce_ta_T, *sce_ta_P;
struct sce_ta_S
{
	unsigned int scta_rcpts_s;	/* #RCPTs sent */
	unsigned int scta_rcpts_r;	/* #replies for RCPTs received */
	unsigned int scta_rcpts_ok;	/* #RCPTs that have been accepted by server */
	int			 scta_rcpt_r;	/* RCPT status */
	unsigned int scta_rcpt_i;	/* index of failed RCPT (iff rcpt_r is set) */
	int			 scta_mail_s;	/* MAIL has been sent? */
	int			 scta_mail_r;	/* MAIL status */
};

/* session context */
typedef struct sce_se_S		sce_se_T, *sce_se_P;
struct sce_se_S
{
	unsigned int	 scse_srv_caps;
	unsigned long	 scse_srv_max_sz_b;
	char			 scse_seid[21];
};

typedef struct sce_sts_S		sce_sts_T, *sce_sts_P;
struct sce_sts_S
{
	st_utime_t	 sces_wr[ST_SE_N];
	st_utime_t	 sces_rd[ST_SE_N];
	st_utime_t	 sces_se;	/* for one session */
	sce_stt_P	 sces_tas;	/* transaction data */
};

static sce_sts_P sc_sts = NULL;
static int stats = 0;

static void
sce_usage(const char *prg)
{
	sm_io_fprintf(smioerr,
		"Usage: %s [options] -r host:port\n"
		"-@ cap      try additional capabilities\n"
#if MTA_USE_TLS
		"            S: STARTTLS\n"
#endif
		"-/ n        if n > 0: srand(n), otherwise some pseudo-random init\n"
		"-0 n        randomly drop connection (random() < n)\n"
		"-1 n        body length\n"
		"-2          flush data before final dot\n"
		"-3          if server greets with 421 sleep 1 second\n"
		"-4 t        stop after t seconds\n"
		"-5          select RCPT per TA, not global\n"
		"-6          use unique addresses (counter)\n"
		"-8          try Deferred RCPT Reply\n"
		"-9          try RCPT status after dot.\n"
		"-A          measure SMTP dialogue times\n"
		"-a stage=code   set action reply code for stage to code\n"
		"   stage:\n"
		"     c: new session (connect)\n"
		"     h: HELO\n"
		"     m: MAIL\n"
		"     r: RCPT\n"
		"     d: DATA\n"
		"     b: Body\n"
		"   code:    action\n"
		"   %06#x   abort session [drop connection]\n"
		"   %06#x   abort transaction [move to next if there is one]\n"
		"   lower 2 bytes can be used as counter\n"
		"-B filename use content of filename for commands to send\n"
		"-b ip       bind to ipv4 address"
		"            (can be specified multiple times)\n"
		"-C n        show counter (print if counter %% n == 0)\n"
		"-D filename[:n] use content of filename for DATA\n"
		"            (can be specified multiple times)\n"
		"            n specifies the relative weight for using filename\n"
		"            file must be in SMTP format but without trailing CRLF.CRLF\n"
		"-d n        set debug level\n"
		"-E          show SMTP dialogue errors\n"
		"-e domain   use domain for EHLO [%s]\n"
		"-F domain   use domain for sender addresses [%s] (can be specified multiple times)\n"
		"-f address  from address (can be specified multiple times)\n"
		"-G          show total rate (requires -m)\n"
		"-g n        number of `teardown' messages (before total)\n"
		"-H ipaddr[_hostname][/n]   issue X-CONNECT ipaddr[ hostname]\n"
		"            (requires server support; can be specified multiple times)\n"
		"            n specifies the relative weight for use\n"
		"-I n        number of `startup' messages\n"
		"-i          interactive display of performance data (requires -C)\n"
		"-l n        data length\n"
		"-M arg      set arg for MAIL\n"
		"-m n        number of messages to send\n"
		"-N          do not use RSET between transactions\n"
		"-n n        number of recipients per transaction\n"
		"-O timeout  I/O timeout [%d].\n"
		"-P          use PIPELINING if offered\n"
		"-p n        use n in addresses as identifier\n"
		"-R address  recipient address (can be specified multiple times)\n"
#if MTA_NETINET6
		"-r address  this can also be a socket specification (including inet6)\n"
#endif
		"-S          stop on errors\n"
		"-s n        total sessions\n"
		"-T n        transactions per thread\n"
		"-t n        concurrent threads\n"
#if MD_HAVE_EPOLL || MD_HAVE_KQUEUE
		"-u          use alternative event system (statethreads)\n"
#endif
		"-w n        wait for n seconds after DATA\n"
		"-W stage=delay   wait for delay seconds before issuing stage command\n"
		"-q n        send n messages with 1..n as body.\n"
		"-Q m        use m as first element in sequence: m..n.\n"
#if MTA_USE_TLS
		"-Xsub                Set STARTTLS options.\n"
		"  ccert              cert\n"
		"  kkey               key\n"
		"  lcipherlist        cipherlist\n"
		"  ooptions           options\n"
#endif
		"-x          do not stop session if there is a transaction error.\n"
		"            if specified twice do not stop the thread either.\n"
		"-Y domain   use domain for recipient addresses [%s]\n"
		"-y rpctarg  RCPT argument (e.g., ORCPT=...)\n"
		"-Z          use SIZE= for MAIL\n"
		"-z          do not read reply to QUIT\n"
		, prg
		, SCE_ST_ABORT_SE
		, SCE_ST_ABORT_TA
		, ehloarg
		, fromdom[0]
		, request_timeout
		, rcptdom[0]
		);
}

static unsigned int
getstage(char arg, bool nextok)
{
	unsigned int stage;

	stage = -1;
	switch (arg) {
	  case 'c':
		stage = STAGE_CONNECT;
		break;
	  case 'h':
	  case 'e':
		stage = STAGE_HELO;
		break;
	  case 'm':
		stage = STAGE_MAIL;
		break;
	  case 'r':
		stage = STAGE_RCPT;
		break;
	  case 'd':
		stage = STAGE_DATA;
		break;
	  case 'b':
		stage = STAGE_DOT;
		break;
	  case 'n':
		if (nextok) {
			stage = STAGE_NEXT;
			break;
		}
		/* fallthrough */
	  default:
		sce_usage(prog);
		exit(EX_USAGE);
	}
	return stage;
}

#define ST_UTIME_MAX ((st_utime_t)-1)

static st_utime_t
next_val(int se, int ta, int stage, int wr)
{
	sce_stt_P sc_stt;

	if (!stats)
		return ST_UTIME_MAX;
	if ((unsigned int) se >= sessions)
		return ST_UTIME_MAX;
	if ((unsigned int)ta >= ta_per_thrd)
		return ST_UTIME_MAX;

	if (ta < 0) {
		if (stage < 0)
			return ST_UTIME_MAX;
		if (stage >= ST_SE_N)
			return sc_sts[se].sces_se;
		if (wr)
			return sc_sts[se].sces_wr[stage];
		else
			return sc_sts[se].sces_rd[stage];
	}

	sc_stt = sc_sts[se].sces_tas;
	if (stage < 0)
		return ST_UTIME_MAX;
	if (stage >= ST_TA_N)
		return sc_sts[se].sces_se;
	if (wr)
		return sc_stt[ta].scet_wr[stage];
	else
		return sc_stt[ta].scet_rd[stage];
	/* NOTREACHED */

	return ST_UTIME_MAX;
}

void
sc_dprintf(int tid, const char *fmt, ...)
{
	va_list ap;

	sm_io_fprintf(smioerr, "[%d:%v] ", tid, st_time());
	va_start(ap, fmt);
	sm_io_vfprintf(smioerr, fmt, ap);
	va_end(ap);
}

void
sc_dprintfs(int tid, const char *seid, const char *fmt, ...)
{
	va_list ap;

	/* maybe don't print seid if it is empty? */
	sm_io_fprintf(smioerr, "[%d:%v:%s] ", tid, st_time(), seid);
	va_start(ap, fmt);
	sm_io_vfprintf(smioerr, fmt, ap);
	va_end(ap);
}


void
sc_dnprintf(int tid, char *buf, size_t bufsz, const char *fmt, ...)
{
	va_list ap;
	int n;

	n = sm_snprintf(buf, bufsz, "[%d:%v] ", tid, st_time());
	if (n > 0 && n < bufsz) {
		va_start(ap, fmt);
		sm_vsnprintf(buf + n, bufsz -n, fmt, ap);
		va_end(ap);
	}
}


static void
some_stats(int se, int ta, int stage, int wr)
{
	unsigned int se_u, ta_u, stage_u, n;
	st_utime_t t_val, t_sum, t_min, t_max;
#if SM_MATH_STATS
	ulong geometric, harmonic;
	double product, inversesum;
#endif

	if (se >= 0 && ta >= 0 && stage >= 0)
		return;

#if SM_MATH_STATS
	product = 1.0;
	inversesum = 0.0;
#endif
	t_val = t_sum = 0;
	t_min = UINT_MAX;
	t_max = 0;
	se_u = ta_u = stage_u = 0;
	if (se >= 0)
		se_u = se;
	if (ta >= 0)
		ta_u = ta;
	if (stage >= 0)
		stage_u = stage;

	n = 0;
	while (t_val >= 0) {
		t_val = next_val(se_u, ta_u, stage_u, wr);
		if (ST_UTIME_MAX == t_val)
			break;
		t_sum += t_val;
		if (t_val < t_min)
			t_min = t_val;
		if (t_val > t_max)
			t_max = t_val;
#if SM_MATH_STATS
		product *= (double) t_val;
		inversesum += (double) 1.0 / (double) t_val;
#endif

		if (se < 0)
			++se_u;
		else if (ta < 0)
			++ta_u;
		else if (stage < 0)
			++stage_u;
		++n;
	}
#if SM_MATH_STATS
	geometric = pow(product, 1.0/(double)n);
	harmonic = (double)n / inversesum;
#endif
	sm_io_fprintf(smioerr,
		"%4d:%4d:%d:%u  min=%4lu, max=%4lu, medium=%ld, geometric=%ld, harmonic=%ld\n"
		, se, ta, stage, n
		, (ulong) t_min, (ulong) t_max
		, (ulong) (t_sum / ((n > 0) ? n : 1))
#if SM_MATH_STATS
		, geometric, harmonic
#else
		, -1, -1
#endif
		);
}

static void
show_stats(void)
{
	unsigned int se, ta, stage;
	sce_stt_P sc_stt;

	if (!stats || NULL == sc_sts)
		return;

	for (se = 0; se < sessions; se++) {
		sc_stt = sc_sts[se].sces_tas;
		for (ta = 0; ta < ta_per_thrd; ta++) {
			for (stage = 0; stage < ST_TA_N; stage++) {
				sm_io_fprintf(smioerr,
					"%4u:%4u:%u  wr=%4lu, rd=%4lu\n"
					, se, ta, stage
					, (ulong) sc_stt[ta].scet_wr[stage]
					, (ulong) sc_stt[ta].scet_rd[stage]
					);
			}
			sm_io_fprintf(smioerr,
				"%4u:%4u    ta=%4lu\n"
				, se, ta
				, (ulong) sc_stt[ta].scet_ta
				);
		}
		for (stage = 0; stage < ST_SE_N; stage++) {
			sm_io_fprintf(smioerr,
				"%4u:    :%u  wr=%4lu, rd=%4lu\n"
				, se, stage
				, (ulong) sc_sts[se].sces_wr[stage]
				, (ulong) sc_sts[se].sces_rd[stage]
				);
		}
		sm_io_fprintf(smioerr,
			"%4u         se=%4lu\n"
			, se
			, (ulong) sc_sts[se].sces_se
			);
	}
	some_stats(-1, 0, 1, 1);
}

static int
read_address(const char *str, sm_sockaddr_P sa)
{
	sm_ret_T ret;
	short port;
	struct hostent *hp;
	sockspec_T sockspec;
	char host[128], *p;

	strlcpy(host, str, sizeof(host));

	ret = parsesockstr(host, &sockspec);
	if (SM_SUCCESS == ret) {
		sockaddr_len_T sock_size;

		ret = sockspec2sock(&sockspec, sa, &sock_size, SS2S_FL_RMUNIX);
		return ret;
	}

	if ((p = strchr(host, ':')) == NULL) {
		sm_io_fprintf(smioerr, "%s: invalid address: %s, missing \":port\"\n", prog, host);
		return EX_USAGE;
	}
	*p++ = '\0';
	port = (short) atoi(p);
	if (port < 1) {
		sm_io_fprintf(smioerr, "%s: invalid port: %s\n", prog, p);
		return EX_USAGE;
	}
	memset(sa, 0, sizeof(*sa));
	sa->sin.sin_family = AF_INET;
	sa->sin.sin_port = htons(port);
	if (host[0] == '\0') {
		sa->sin.sin_addr.s_addr = INADDR_ANY;
		return 0;
	}
	sa->sin.sin_addr.s_addr = inet_addr(host);
	if (INADDR_NONE == sa->sin.sin_addr.s_addr) {
		/* not dotted-decimal */
		if ((hp = gethostbyname(host)) == NULL) {
			sm_io_fprintf(smioerr, "%s: can't resolve address: %s\n", prog, host);
			return EX_OSERR;
		}
		memcpy(&sa->sin.sin_addr, hp->h_addr, hp->h_length);
	}
	return 0;
}


static void
sc_initp(void)
{
#if MTA_USE_TLS
	sm_ret_T ret;

	if (!try_starttls)
		return;

	ret = sm_tlsversionok();
	if (sm_is_err(ret))
		err_quit(errfd, "ERROR: sm_tlsversionok=%#x", ret);
	tls_ctx = NULL;
	ret = sm_tls_init_library(&tls_ctx, false);
	if (sm_is_err(ret))
		err_quit(errfd, "ERROR: sm_tls_init_library=%#x", ret);

	TLS_GEN_PATHS(tls_cnf, NULL);
#if MTA_TLS_DEBUG
	tls_cnf.tlscnf_debug = 14;
#endif

	ret = sm_tls_init(tls_ctx, TLS_I_CLT, false, &tls_cnf);
	if (sm_is_err(ret))
		err_quit(errfd, "ERROR: sm_tls_init=%#x", ret);
	ret = sm_set_tls_log(tls_ctx, smioerr, SMIOERR_FILENO, 14);
	if (sm_is_err(ret))
		err_report(errfd, "ERROR: sm_set_tls_log=%#x", ret);

	return;

 enomem:	/* used by TLS_GEN_PATHS() */
	err_quit(errfd, "ERROR: sm_tls_init=ENOMEM");
#endif /* MTA_USE_TLS */
}

int
main(int argc, char *argv[])
{
	extern char    *optarg;
	int             opt, n, threads;
	int             raddr;
	unsigned int    u, off, stage;
	long int        tc;
	size_t		j;
	ssize_t		bytesread;
	sce_stt_P	sc_stt;
	char *str;

	prog = argv[0];
	if (getuid() == 0 || geteuid() == 0) {
		sm_io_fprintf(smioerr, SM_DONTRUNASROOT "\n", prog);
		exit(EX_USAGE);
	}
	raddr = 0;
	sessions = ta_per_thrd = threads = 1;
	rcpt[0] = from[0] = NULL;
	bytesread = 0;
	fromdom[0] = "local.host";
	rcptdom[0] = "local.host";
#if MTA_USE_TLS
	sm_memzero(&tls_cnf, sizeof(tls_cnf));
#endif

	/* Parse arguments; avail: 7LUvVz */
	while ((opt = getopt(argc, argv,
			"@:/:0:1:234:5689Aa:B:b:C:c:D:d:Ee:F:f:Gg:H:hI:il:M:m:Nn:O:o:Pp:Q:q:R:r:Ss:T:t:uW:w:X:xY:y:Zz"))
				!= -1)
	{
		switch (opt) {
		  case '@':
			for (u = 0; u < strlen(optarg); u++) {
				opt = (int) optarg[u];
				switch (opt) {
#if MTA_USE_TLS
				  case 'S': try_starttls = true; break;
#endif
				  default:
					err_quit_c(errfd, EX_USAGE, "ERROR: unknown -@ option %c", (char) opt);
					break;
				}
			}
			break;
		   case '/':
			n = atoi(optarg);
			if (n > 0)
				srand(n);
			else
				srand(getpid() + time(NULL));
			break;
		  case '0':
			random_drop = atoi(optarg);
			break;
		  case '1':
			datalen = (unsigned long) strtoul(optarg, NULL, 0) + strlen(MAIL_HEADER);
			break;
		  case '2':
			flushdata = 1;
			break;
		  case '3':
			delayonerror = true;
			break;
		  case '4':
			stoptime = st_time() + strtoul(optarg, NULL, 0);
			break;
		  case '5':
			rpct_per_ta = true;
			break;
		  case '6':
			uniqueaddrs = true;
			break;
		  case '8':
			try_drr = true;
			break;
		  case '9':
			try_rsad = true;
			break;
		  case 'A':
			stats = 1;
			break;
		  case 'a':
			if (optarg != NULL && ISALPHA(optarg[0]) && optarg[1] == '=')
				off = 2;
			else {
				sce_usage(prog);
				break;
			}

			u = (unsigned int) strtoul(optarg + off, NULL, 0);
			stage = getstage(optarg[0], false);
			SM_ASSERT(stage < STAGES);
			codes[stage] = u;
			break;
		  case 'B':
			cmdfile = strdup(optarg);
			if (NULL == cmdfile) {
				sm_io_fprintf(smioerr, "%s: strdup(%s)=NULL\n", prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'b':
			if (nclt_ips >= MAX_CLT_IPS - 1) {
				sm_io_fprintf(smioerr, "%s: too many client IP addresses (%d) max\n", prog, MAX_CLT_IPS);
				break;
			}
			clt_ip = inet_addr(optarg);
			if (clt_ip == INADDR_NONE) {
				sm_io_fprintf(smioerr, "%s: bogus client IP address \"%s\"\n", prog, optarg);
				break;
			}
			clt_ips[nclt_ips++] = clt_ip;
			use_bind = true;
			break;
		  case 'C':
			count = (unsigned long) strtoul(optarg, NULL, 0);
			break;
		  case 'T':
		  case 'c':
			ta_per_thrd = (unsigned int) strtoul(optarg, NULL, 0);
			if (ta_per_thrd < 1) {
				sm_io_fprintf(smioerr,
					"%s: invalid number of transactions per thread: %s\n",
					prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'd':
			debug = atoi(optarg);
			if (debug < 0) {
				sm_io_fprintf(smioerr, "%s: invalid number for debug: %s\n",
					prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'D':
			if (ndatafiles >= MAXDF - 1) {
				sm_io_fprintf(smioerr, "%s: too many datafiles (%d) max\n",
					prog, MAXDF);
				break;
			}
			if ((str = strchr(optarg, ':')) != NULL) {
				*str++ = '\0';
				use_df_perc = true;
				df_perc[ndatafiles] = atoi(str);
			}
			datafile = strdup(optarg);
			if (NULL == datafile) {
				sm_io_fprintf(smioerr, "%s: strdup(%s)=NULL\n", prog, optarg);
				exit(EX_OSERR);
			}
			datafiles[ndatafiles++] = datafile;
			break;
		  case 'E':
			++showerror;
			break;
		  case 'e':
			ehloarg = strdup(optarg);
			if (NULL == ehloarg) {
				sm_io_fprintf(smioerr, "%s: strdup(%s)=NULL\n", prog, optarg);
				exit(EX_USAGE);
			}
			/*
			 *  hack because I can't figure out how to pass [...]
			 *  unmolested by sh through various sh scripts.
			 */
			if ('\'' == ehloarg[0]) {
				j = strlen(ehloarg);
				if (j > 2 && '\'' == ehloarg[j - 1]) {
					ehloarg[j - 1] = '\0';
					++ehloarg;
				}
			}
			break;
		  case 'f':
			if (fromaddrs >= SC_MAXADDRS - 1) {
				sm_io_fprintf(smioerr,
					"%s: too many addresses=%d, max=%d\n",
					prog, fromaddrs, SC_MAXADDRS);
				exit(EX_USAGE);
			}
			from[fromaddrs++] = optarg;
			break;
		  case 'F':
			if (fromdoms >= SC_MAXADDRS - 1) {
				sm_io_fprintf(smioerr,
					"%s: too many domains=%d, max=%d\n",
					prog, fromdoms, SC_MAXADDRS);
				exit(EX_USAGE);
			}
			fromdom[fromdoms++] = optarg;
			break;
		  case 'G':
			show_rate = 1;
			break;
		  case 'g':
			teardown = (unsigned long) atol(optarg);
			break;
		  case 'H':
			if (nhostinfos >= MAXHI - 1) {
				sm_io_fprintf(smioerr, "%s: too many hostinfo options, max=%d\n", prog, MAXHI);
				break;
			}
			if (NULL == optarg) {
				sm_io_fprintf(smioerr, "%s: missing arg for -%c\n", prog, opt);
				exit(EX_USAGE);
			}
			if ((str = strchr(optarg, '/')) != NULL) {
				*str++ = '\0';
				use_hi_perc = true;
				hi_perc[nhostinfos] = atoi(str);
			}
			hostinfo = strdup(optarg);
			if (NULL == hostinfo) {
				sm_io_fprintf(smioerr, "%s: strdup(%s)=NULL\n", prog, optarg);
				exit(EX_OSERR);
			}
			tr_simple(hostinfo, "_", " ");
			hostinfos[nhostinfos++] = hostinfo;
			break;
		  case 'I':
			startup = (unsigned long) atol(optarg);
			break;
		  case 'i':
			idisplay = 1;
			break;
		  case 'M':
			if (NULL == optarg) {
				sm_io_fprintf(smioerr, "%s: missing arg for -%c\n", prog, opt);
				exit(EX_USAGE);
			}
			mailarg = strdup(optarg);
			if (NULL == mailarg) {
				sm_io_fprintf(smioerr, "%s: strdup(%s) failed\n", prog, optarg);
				exit(EX_OSERR);
			}
			/* Warning: this means we can't use "_" in args! */
			tr_simple(mailarg, "_", " ");
			break;
		  case 'm':
			ta_total = (unsigned long) atol(optarg);
			break;
		  case 'l':
			datalen = (unsigned long) strtoul(optarg, NULL, 0);
			break;
		  case 'N':
			rset = 0;
			break;
		  case 'n':
			rcpts = (unsigned int) atoi(optarg);
			if (rcpts < 1) {
				sm_io_fprintf(smioerr, "%s: invalid number of rcpts: %s\n",
					prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'O':
			request_timeout  = atoi(optarg);
			if (request_timeout <= 0) {
				sm_io_fprintf(smioerr, "%s: invalid timeout: %s\n",
					prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'P':
			use_pipelining = true;
			break;
		  case 'p':
			postfix = atoi(optarg);
			break;
		  case 'r':
			if ((n = read_address(optarg, &rmt_addr)) != 0) {
				sm_io_fprintf(smioerr, "%s: invalid remote address: %s\n",
					prog, optarg);
				exit(n);
			}
			raddr = 1;
			break;
		  case 'R':
			if (rcptaddrs >= SC_MAXADDRS - 1) {
				sm_io_fprintf(smioerr, "%s: too many addresses=%d, max=%d\n",
					prog, rcptaddrs, SC_MAXADDRS);
				exit(EX_USAGE);
			}
			rcpt[rcptaddrs++] = optarg;
			break;
		  case 's':
			sessions = (unsigned int) atoi(optarg);
			if (sessions < 1) {
				sm_io_fprintf(smioerr, "%s: invalid number of sessions: %s\n",
					prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'S':
			stoponerror++;
			break;
		  case 't':
			threads = atoi(optarg);
			if (threads < 1) {
				sm_io_fprintf(smioerr, "%s: invalid number of threads: %s\n",
					prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'q':
			sequence = atoi(optarg);
			if (sequence < 1) {
				sm_io_fprintf(smioerr, "%s: invalid number for sequence: %s\n"
					, prog, optarg);
				exit(EX_USAGE);
			}
			break;
		  case 'Q':
			seqfirst = atoi(optarg);
			if (seqfirst < 0) {
				sm_io_fprintf(smioerr,
					"%s: invalid number for sequence begin: %s\n"
					, prog, optarg);
				exit(EX_USAGE);
			}
			break;
#if MD_HAVE_EPOLL || MD_HAVE_KQUEUE
		  case 'u':
			alteventsys = ST_EVENTSYS_ALT;
			break;
#endif
		  case 'W':
			if (optarg != NULL && ISALPHA(optarg[0]) && optarg[1] == '=')
				off = 2;
			else {
				sce_usage(prog);
				break;
			}

			u = (unsigned int) strtoul(optarg + off, NULL, 0);
			stage = getstage(optarg[0], true);
			SM_ASSERT(stage < STAGES);
			waits[stage] = u;
			break;
		  case 'w':
			waitdata = atoi(optarg);
			break;
#if MTA_USE_TLS
		  case 'X':
			opt = (int) optarg[0];
			switch (opt) {
			  case 'c': tls_cnf.tlscnf_cert = strdup(optarg + 1); break;
			  case 'k': tls_cnf.tlscnf_key = strdup(optarg + 1); break;
			  case 'l': tls_cnf.tlscnf_cipherlist = strdup(optarg + 1); break;
			  case 'o': tls_cnf.tlscnf_options = strtoul(optarg + 1, NULL, 0);
					tls_cnf.tlscnf_options_isset = 1;
					break;
			  default:
				err_quit_c(errfd, EX_USAGE, "ERROR: unknown -Y option %c", (char) opt);
				break;
			}
			break;
#endif
		  case 'x':
			++contiftaerror;
			break;
		  case 'Y':
			if (rcptdoms >= SC_MAXADDRS - 1) {
				sm_io_fprintf(smioerr, "%s: too many domains=%d, max=%d\n",
					prog, rcptdoms, SC_MAXADDRS);
				exit(EX_USAGE);
			}
			rcptdom[rcptdoms++] = optarg;
			break;
		  case 'y':
			orcpt = optarg;
			break;
		  case 'Z':
			usesize = true;
			break;
		  case 'z':
			dontread_QUIT_reply = true;
			break;
		  case 'h':
		  case '?':
			sce_usage(prog);
			exit(EX_USAGE);
		}
	}
	if (!raddr) {
		sm_io_fprintf(smioerr, "%s: remote address required\n", prog);
		exit(EX_USAGE);
	}
	if (use_df_perc) {
		for (n = 0; n < ndatafiles; n++)
			df_total += df_perc[n];
	}
	if (use_hi_perc) {
		for (n = 0; n < nhostinfos; n++)
			hi_total += hi_perc[n];
	}

	/* number of recipients not set? */
	if (0 == rcpts) {
		if (rcptaddrs > 1)
			rcpts = rcptaddrs;
		else
			rcpts = 1;
	}
	if (0 == fromdoms)
		fromdoms = 1;
	if (0 == rcptdoms)
		rcptdoms = 1;

	if (0 == bytesread) {
		j = strlcpy((char *)databuf, MAIL_HEADER, sizeof(databuf));
		for (j = strlen(MAIL_HEADER); j < sizeof(databuf); j++)
			databuf[j] = ' ' + (j % 64);
		for (j = 80; j < sizeof(databuf) - 1; j += 80) {
			databuf[j] = '\r';
			databuf[j + 1] = '\n';
		}
	}

	tc = (long) sessions * (long) ta_per_thrd;
	if (debug)
		sm_io_fprintf(smioerr, "%s: starting client [%d]\n", prog, threads);

	if (stats) {
		j = sessions * sizeof(*sc_sts);
		if (j < sessions || j < sizeof(*sc_sts)) {
			sm_io_fprintf(smioerr, "%s: statistics too large %lu\n", prog, j);
			exit(EX_OSERR);
		}
		sc_sts = malloc(j);
		if (NULL == sc_sts) {
			sm_io_fprintf(smioerr, "%s: cannot allocate statistics %lu\n",
				prog, j);
			exit(EX_OSERR);
		}
		sm_memzero(sc_sts, j);

		j = tc * sizeof(*sc_stt);
		if (j < (size_t) tc || j < sizeof(*sc_stt)) {
			sm_io_fprintf(smioerr, "%s: statistics too large (TA) %lu\n",
				prog, j);
			exit(EX_OSERR);
		}
		sc_stt = malloc(j);
		if (NULL == sc_stt) {
			sm_io_fprintf(smioerr, "%s: cannot allocate statistics (TA) %lu\n",
				prog, j);
			exit(EX_OSERR);
		}
		sm_memzero(sc_stt, j);
		off = 0;
		for (u = 0; u < sessions; u++) {
			sc_sts[u].sces_tas = &(sc_stt[off]);
			off += ta_per_thrd;
		}
	}

	sc_initp();

	if (alteventsys && st_set_eventsys(ST_EVENTSYS_ALT) < 0)
		print_sys_error("WARN: can't switch to alternate event system");

	/* Initialize the ST library */
	if (st_init() < 0) {
		print_sys_error("st_init");
		exit(EX_OSERR);
	}
	if (idisplay)
		sm_io_fprintf(smioerr,
				"   msg    time_i   time_t  msgs/s t-msgs/s  busy  conc\n");

	for (n = 0; n < threads; n++) {
		if (debug)
			sm_io_fprintf(smioerr, "%s: starting client %d/%d\n",
				prog, n, threads);
		if (st_thread_create(handle_request, (void *) n, 0, 0) == NULL) {
			print_sys_error("st_thread_create");
			exit(EX_OSERR);
		}
	}

	/* wait for them... */
	st_sleep(1);
	while (busy > 0)
		st_sleep(1);

	if (idisplay) {
		sm_io_putc(smioerr, '\n');
		sm_io_flush(smioerr);
	}
	if (show_rate) {
		if (!total_rate_set && firstmail > 0) {
			time_t elapsed;

			elapsed = st_time() - firstmail;
			if (elapsed > 0)
				 total_rate = (ta_cnt - startup_act) / elapsed;
		}
		sm_io_fprintf(smioerr,
			"%s: total=%lu, total_rate=%lu (tc=%ld, ta=%ld)\n",
			prog, ta_cnt, total_rate, tc, ta_total);
	}
	else {
		sm_io_fprintf(smioerr,
			"%s: total=%lu (should be %lu/%lu)\n",
			prog, ta_cnt, tc, ta_total);
	}
	if (use_df_perc) {
		for (n = 0; n < ndatafiles; n++) {
		sm_io_fprintf(smioerr, "n=%d, df_used=%d, perc=%d, df_perc=%d\n"
			, n, df_used[n], df_used[n] * df_total / ta_cnt
			, df_perc[n]);
		}
	}
	show_stats();
	return 0;
}


/* before changing these, check the macros! */
#define SMTP_OK		0
#define SMTP_NONE	1	/* no SMTP reply (PIPELINING) */
#define SMTP_CONT	2	/* 3xy */
#define SMTP_AN		3	/* SMTP reply type isn't 2 or 3 */
#define SMTP_SSD	4	/* 421 */
#define SMTP_RD		10	/* read error */
#define SMTP_WR		20	/* write error */
#define SMTP_IS_OK(r)	((r) <= SMTP_CONT)
#define SMTP_IS_CONT(r)	((r) == SMTP_CONT)
#define SMTP_IO_ERR(r)	((r) >= SMTP_RD)
#define SMTP_FATAL(r)	((r) >= SMTP_SSD)

static int
smtpwrite(unsigned long l, sm_file_T *fp, int tid, int i)
{
	size_t wr;
	sm_ret_T ret;
	ssize_t b;

	while (l > 0) {
		wr = SM_MIN(l, sizeof(databuf));
		ret = sm_io_write(fp, databuf, wr, &b);
		if ((int) b != wr || ret != SM_SUCCESS) {
			sc_dprintf(tid,
				"data write error i=%d, n=%lu, r=%d, ret=%#x\n"
			    , i, l, (int) b, ret);
			return SMTP_WR;
		}
		l -= wr;
	}
	return SMTP_OK;
}

static int
smtpdata(unsigned long l, sm_file_T *fp, int tid, int i)
{
	int r;

	if (datafile != NULL) {
		sm_file_T *dfp;
		ssize_t bytesread;
		unsigned int prev;

		prev = ndatafile;
		if (++ndatafile >= ndatafiles)
			ndatafile = 0;
		if (use_df_perc) {
			unsigned long tas;

			tas = ta_cnt > 0 ? ta_cnt : 1;
			while (ndatafile != prev) {
				if (df_used[ndatafile] * df_total / tas < df_perc[ndatafile])
					break;
				if (++ndatafile >= ndatafiles)
					ndatafile = 0;
			}
		}
		datafile = datafiles[ndatafile];
		df_used[ndatafile]++;
		r = sm_io_open(SmStStdio, datafile, SM_IO_RDONLY, &dfp,
			SM_IO_WHAT_END);
		if (r != SM_SUCCESS) {
			sm_io_fprintf(smioerr, "sm_io_open(%s)=%r\n", datafile, r);
			return r;
		}
		do {
			r = sm_io_read(dfp, databuf, sizeof(databuf), &bytesread);
			if (SM_IO_EOF == r) {
				r = SMTP_OK;
				break;
			}
			if (sm_is_err(r))
				break;
			r = smtpwrite(bytesread, fp, tid, i);
			if (flushdata)
				(void) sm_io_flush(fp);
		} while (SMTP_OK == r);
		sm_io_close(dfp, SM_IO_CF_NONE);
	}
	else
		r = smtpwrite(l, fp, tid, i);
	return r;
}

/*
**  RD_SRV_CAP -- read server capabilities
**
**	Parameters:
**		ehlo_reply -- EHLO reply text
**		sc_se -- SMTPC session context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
rd_srv_cap(sm_str_P ehlo_reply, sce_se_P sc_se)
{
	char *opt;

	/* "250 " + something useful: at least 4 characters */
	if (sm_str_getlen(ehlo_reply) < 8)
		return SM_SUCCESS;
	if (sm_str_rd_elem(ehlo_reply, 0) != '2' ||
	    sm_str_rd_elem(ehlo_reply, 1) != '5' ||
	    sm_str_rd_elem(ehlo_reply, 2) != '0' ||
	    (sm_str_rd_elem(ehlo_reply, 3) != ' ' &&
	     sm_str_rd_elem(ehlo_reply, 3) != '-'))
		return SM_SUCCESS;	/* error?? */

#define SMTPC_IS_CAPN(str)	(sm_strcaseeqn(opt, (str), strlen(str)))
#define SMTPC_IS_CAP(str)	(sm_strcaseeq(opt, (str)))

	opt = (char *) sm_str_getdata(ehlo_reply);
	SM_REQUIRE(opt != NULL);
	opt += SMTP_REPLY_OFFSET;

	if (SMTPC_IS_CAP("pipelining"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_PIPELINING);
	else if (SMTPC_IS_CAP("8bitmime"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_8BITMIME);
	else if (SMTPC_IS_CAP("size"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_SIZE);
	else if (SMTPC_IS_CAPN("size ")) {
		ulonglong_T max_sz_b;	/* off_t? */
		char *endptr;

		errno = 0;

		/* 5 == strlen("size ") */
		max_sz_b = sm_strtoull(opt + 5, &endptr, 10);
		if (!((ULLONG_MAX == max_sz_b && ERANGE == errno) ||
		      endptr == opt + 5))
		{
			SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_SIZE);
			sc_se->scse_srv_max_sz_b = max_sz_b;
		}
		/* else: silently ignore ... */
	}
	else if (SMTPC_IS_CAPN("seid ")) {
		/* 5 == strlen("seid ") */
		strlcpy(sc_se->scse_seid, opt + 5, sizeof(sc_se->scse_seid));
	}
	else if (SMTPC_IS_CAP("enhancedstatuscodes"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_ENHSTAT);
	else if (SMTPC_IS_CAPN("auth "))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_AUTH);
	else if (SMTPC_IS_CAP("starttls"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_STARTTLS);
	else if (SMTPC_IS_CAP("drr"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_DRR);
	else if (SMTPC_IS_CAP("prdr"))
		SCSE_SET_CAP(sc_se->scse_srv_caps, SCSE_CAP_RSAD);
	return SM_SUCCESS;
}

/*
**  SMTPCOMMAND -- send one SMTP command and read reply
**
**	Parameters:
**		str -- SMTP command
**		l -- length of str
**		fp -- filepointer for output
**		out -- reply from server (to be filled in)
**		tid -- thread id
**		cta -- transaction counter
**		stage -- SMTP stage
**		t_write -- time for write (output)
**		t_read -- time for read (output)
**		sc_se -- session context
**		sc_ta -- transaction context
**
**	Returns:
**		SMTP_* (see above)
*/

static int
smtpcommand(char *str, int l, sm_file_T *fp, sm_str_P out, int tid, unsigned int cta, int stage, st_utime_t *t_write, st_utime_t *t_read, sce_se_P sc_se, sce_ta_P sc_ta)
{
	sm_ret_T        ret;
	ssize_t         b;
	st_utime_t	t0, t1;
	bool			used;
	unsigned int	rcpts_ad;

	/* avoid bogus "might be used uninitialized" warnings */
	t0 = t1 = 0;
	if (stage >= 0 && stage <= STAGES && waits[stage] > 0)
		sleep(waits[stage]);
	if (debug > 3) {
		sc_dprintfs(tid, sc_se->scse_seid, "send: ");
		sm_io_write(smioerr, (uchar *) str, l, &b);
		sm_io_flush(smioerr);
	}
	if (stats)
		t0 = st_utime();
	ret = sm_io_write(fp, (uchar *) str, l, &b);
	if (b != l) {
		sc_dprintfs(tid, sc_se->scse_seid,
			"write error cta=%u, n=%d, r=%d, ret=%#x\n",
			cta, l, (int) b, ret);
		if (stoponerror)
			sessions = 0;
		return SMTP_WR;
	}
	if (!SC_PIPELINE_OK(stage)
	    || !SCSE_IS_CAP(sc_se->scse_srv_caps, SCSE_CAP_PIPELINING))
	{
		ret = sm_io_flush(fp);
		if (sm_is_error(ret)) {
			sc_dprintfs(tid, sc_se->scse_seid,
				"flush error cta=%u, n=%d, ret=%#x\n",
				cta, (int) b, ret);
			if (stoponerror)
				sessions = 0;
			return SMTP_WR;
		}
	}
	if (stats && t_write != NULL) {
		t1 = st_utime();
		*t_write = t1 - t0;
		t0 = t1;
	}
	if (STAGE_RCPT == stage)
		++sc_ta->scta_rcpts_s;

	rcpts_ad = 0;

	/* also check how much data has been sent? */
	if (SC_PIPELINE_OK(stage)
	    && SCSE_IS_CAP(sc_se->scse_srv_caps, SCSE_CAP_PIPELINING)
	    && !sm_io_getinfo(fp, SM_IO_IS_READABLE, NULL))
	{
		return SMTP_NONE;
	}
	if (STAGE_QUIT == stage && dontread_QUIT_reply)
		return SMTP_NONE;

  again:
	used = false;
	do {
		time_t before, after;

		sm_str_clr(out);
#if DEBUG
		before = st_time();
#else
		before = 0;
#endif
		ret = sm_fgetline0(fp, out);
#if DEBUG
		after = st_time();
#else
		after = 0;
#endif

		if (debug > 3) {
			sc_dprintfs(tid, sc_se->scse_seid, "rcvd [len=%d, res=%#x]: ",
			      sm_str_getlen(out), ret);
			sm_io_write(smioerr, sm_str_getdata(out), sm_str_getlen(out), &b);
			sm_io_putc(smioerr, '\n');
			sm_io_flush(smioerr);
		}
		if (sm_is_error(ret)) {
			if (STAGE_QUIT != stage) {
				int cmdl;

				cmdl = l > 2 ? l - 2 : l;
				sc_dprintfs(tid, sc_se->scse_seid,
					"cmd=%.*s, error=read, cta=%u, n=%d, ret=%#x"
#if DEBUG
					", after(%ld)-before(%ld)=%ld\n"
#else
					"\n"
#endif
					, cmdl, str, cta, (int) b, ret
#if DEBUG
					, (long) after, (long) before, (long) (after - before)
#endif
					);
			}
			if (stoponerror)
				sessions = 0;
			return SMTP_RD;
		}
		if (sm_str_getlen(out) == 0 ||
		    (sm_str_rd_elem(out, 0) != '2' && sm_str_rd_elem(out, 0) != '3'))
		{
			ret = SMTP_AN;
			if (use_pipelining &&
			    SCSE_IS_CAP(sc_se->scse_srv_caps, SCSE_CAP_PIPELINING) &&
			    (STAGE_RCPT == stage || STAGE_DATA == stage))
				break;
			else if (STAGE_DOT == stage && (use_rsad || use_drr))
				break;
			else
				return ret;
		}
		if (STAGE_HELO == stage)
			(void) rd_srv_cap(out, sc_se);
	} while (!sm_is_error(ret) && sm_str_getlen(out) > 4 &&
		 sm_str_rd_elem(out, 3) == '-');

	if (stats && t_read != NULL) {
		t1 = st_utime();
		*t_read = t1 - t0;
	}

	if (sm_str_getlen(out) > 3
	    && sm_str_rd_elem(out, 0) == '4'
	    && sm_str_rd_elem(out, 1) == '2'
	    && sm_str_rd_elem(out, 2) == '1')
	{
		return SMTP_SSD;
	}

	if (sm_str_getlen(out) > 3
	    && sm_str_rd_elem(out, 0) != '2' && sm_str_rd_elem(out, 0) != '3')
		ret = SMTP_AN;
	else if (sm_str_getlen(out) > 3 && sm_str_rd_elem(out, 0) == '3')
		ret = SMTP_CONT;
	else
		ret = SMTP_OK;	/* not really correct but good enough */

	if (use_pipelining && SCSE_IS_CAP(sc_se->scse_srv_caps, SCSE_CAP_PIPELINING) &&
	    (STAGE_RCPT == stage || STAGE_DATA == stage
	     /* || STAGE_LDAT == stage */ ))
	{
		if (SMTP_NONE == sc_ta->scta_mail_r) {
			sc_ta->scta_mail_r = ret;
			used = true;
		}
		else if (sc_ta->scta_rcpts_r < sc_ta->scta_rcpts_s) {
			if (SMTP_OK == ret)
				++sc_ta->scta_rcpts_ok;
			else {
				sc_ta->scta_rcpt_r = ret;
				sc_ta->scta_rcpt_i = sc_ta->scta_rcpts_r;
			}

			++sc_ta->scta_rcpts_r;
#if 0
			if (sc_ta->scta_rcpts_r == sc_ta->scta_rcpts_s)
				/* done; what to do? */;
#endif
			used = true;
		}

		if (!SMTP_IS_OK(ret))
			sm_str_cpy(reply_str, out) ;

		/* need to collect all replies? */
		if (STAGE_DATA == stage && used)
			goto again;
	}

	if (STAGE_DOT == stage && use_rsad) {
			if (0 == rcpts_ad) {
				if (ret != SMTP_CONT)
					rcpts_ad = sc_ta->scta_rcpts_ok + 1;	/* avoid "looping" */
			}
			if (!SMTP_IS_OK(ret) && showerror && rcpts_ad <= sc_ta->scta_rcpts_ok)
					sm_io_fprintf(smioerr,
						"RCPT=error, text=%@T, se=%d, ta=%u, rcpt=%u\n",
						out, sessions, rcpts_ad);
			++rcpts_ad;
			if (rcpts_ad <= sc_ta->scta_rcpts_ok + 1)
				goto again;
	}

	if (STAGE_DOT == stage && use_drr) {
			++rcpts_ad;
			if (rcpts_ad <= sc_ta->scta_rcpts_ok)
				goto again;
	}

	if (STAGE_MAIL == stage && SMTP_OK == ret)
		sc_ta->scta_mail_r = ret;

	return ret;
}

static int
smtpcmds(sm_file_T *fp, int tid, sm_str_P out)
{
	int r;
	sm_file_T *cfp;
	sce_ta_T sc_ta;
	sce_se_T sc_se;

	r = sm_io_open(SmStStdio, cmdfile, SM_IO_RDONLY, &cfp, SM_IO_WHAT_END);
	if (r != SM_SUCCESS) {
		sm_io_fprintf(smioerr, "sm_io_open(%s)=%r\n", cmdfile, r);
		return r;
	}

	memset(&sc_se, 0, sizeof(sc_se));
	/* should be done at each transaction. */
	memset(&sc_ta, 0, sizeof(sc_ta));
	do {
		sm_str_clr(out);
		r = sm_fgetline(cfp, out);
		if (SM_IO_EOF ==r) {
			r = SM_SUCCESS;
			break;
		}
		if (sm_is_err(r))
			break;

		/* this doesn't handle the DATA section! */
		r = smtpcommand((char *)sm_str_getdata(out), sm_str_getlen(out),
				fp, out, tid, 0, -1, NULL, NULL, &sc_se, &sc_ta);
		if (!SMTP_IS_OK(r) && showerror)
			sm_io_fprintf(smioerr, "text=%@T\n", out);
	} while (!SMTP_IO_ERR(r) && !SMTP_FATAL(r));
	sm_io_close(cfp, SM_IO_CF_NONE);
	return r;
}

static int
clt_bind(int sock)
{
	sm_sockaddr_T clt_addr;
	int addrlen;

	addrlen = sizeof(clt_addr.sin);
	memset(&clt_addr, 0, sizeof(clt_addr));
	clt_addr.sa.sa_family = AF_INET;
	/* clt_addr.sin.sin_port = htons(port); */

	/* just round robin? maybe random or percentage? */
	if (++nclt_ip >= nclt_ips)
		nclt_ip = 0;
	clt_addr.sin.sin_addr.s_addr = clt_ips[nclt_ip];
	clt_addr.sa.sa_family = AF_INET;
	if (bind(sock, &clt_addr.sa, addrlen) < 0)
		return errno;
	return SM_SUCCESS;
}

/*
 *  how to handle an error?
 *  warning: this is currently "flow through" code!
 *  there should be a mode to continue which would allow us
 *  to "try" again after an error without decreasing the concurrency
 *  (which prevents us from testing the "recovering" capabilities
 *  of the system).
 */
#define HDL_ERR(delay)	\
	if (contiftaerror > 1) {	\
		if (delay && delayonerror) st_sleep(1);	\
		goto next_se;	\
	}	\
	goto fail

#define HDL_ERR2(where, r)	\
		if (showerror)	\
			sc_dprintfs(tid, sc_se.scse_seid,	\
				"%s=error, text=%@T, se=%d, ta=%u, r=%d\n",	\
				where, out, sessions, cta, r);	\
		if (stoponerror)	\
			sessions = 0;	\
		if (SMTP_IO_ERR(r) && contiftaerror > 1)	\
			goto next_se;	\
		if (contiftaerror)	\
			goto next_ta;	\
		break

static void *
handle_request(void *arg)
{
	st_netfd_t      rmt_nfd;
	int             sock, n, tid, r, rok, myseq;
	unsigned int    cta, j;
	ssize_t         b;
	sm_file_T      *fp;
	sm_str_P        out;
	sm_ret_T        ret;
	char		*s, buf[SM_LINE_LEN], sbuf[64];
	st_utime_t	t0, t1, ts0;
	st_utime_t	*pt0, *pt1;
	sce_sts_P	sc_stse;
	sce_stt_P	sc_stt;
	sce_stt_P	sc_stta;
	sce_ta_T	sc_ta;
	sce_se_T	sc_se;
	bool		incrcc;
#if MTA_USE_TLS
	sm_file_T      *fptls = NULL;
	SSL *se_con = NULL;
	tlsi_ctx_P se_tlsi = NULL;
	bool tlsused = false;
#endif

	++busy;
	tid = (int) arg;

	/* avoid bogus "might be used uninitialized" warnings */
	t0 = t1 = ts0 = 0;
	pt0 = NULL;
	pt1 = NULL;
	sc_stt = NULL;
	sc_stta = NULL;
	sc_stse = NULL;
	incrcc = false;
	fp = NULL;
	sc_se.scse_seid[0] = '\0';

	if (debug)
		sc_dprintf(tid, "client: transactions=%u\n", ta_per_thrd);

	out = sm_str_new(NULL, SM_LINE_LEN, SM_IOBUFSIZE);
	if (NULL == out) {
		sm_snprintf(buf, sizeof(buf), "[%d] str new failed", tid);
		print_sys_error(buf);
		if (stoponerror)
			sessions = 0;
		goto done;
	}

	if (0 == firstNmail)
		firstNmail = st_time();
	if (0 == firstmail && (startup == 0 || ta_cnt >= startup)) {
		firstmail = st_time();
		startup_act = ta_cnt;
	}

	if (use_pipelining) {
		reply_str = sm_str_new(NULL, SM_LINE_LEN, SM_IOBUFSIZE);
		if (NULL == reply_str) {
			sm_snprintf(buf, sizeof(buf), "[%d] str new failed", tid);
			print_sys_error(buf);
			if (stoponerror)
				sessions = 0;
			goto done;
		}
	}
	else
		reply_str = NULL;


	/*
	**  Only run a certain number of sessions;
	**  this is global variable but we can manipulate it without locking.
	*/

	while (sess_cnt < sessions) {
		if (stoptime > 0 && st_time() > stoptime && sessions > 0) {
			sm_io_fprintf(smioerr, "stopped because of timelimit %ld\n", (long) stoptime);
			sessions = 0;
			break;
		}
		incrcc = false;
		if (stats && sc_sts != NULL) {
			sc_stse = &(sc_sts[sess_cnt]);
			sc_stt = sc_sts[sess_cnt].sces_tas;
		}
		++sess_cnt;

		if (0 == firstmail && (startup == 0 || ta_cnt >= startup)) {
			firstmail = st_time();
			startup_act = ta_cnt;
		}
		myseq = sequence;
		if (debug)
			sc_dprintf(tid,
				"client: myseq=%d, sequence=%d, seqfirst=%d\n"
				, myseq, sequence, seqfirst);
		if (myseq == seqfirst) {
			sc_dprintf(tid, "done: myseq=%d, seqfirst=%d\n", myseq, seqfirst);
			goto done;
		}
		if (sequence > seqfirst)
			--sequence;

		if (stats)
			ts0 = t0 = st_utime();

		/* Connect to remote host */
		if ((sock = socket(rmt_addr.sa.sa_family, SOCK_STREAM, 0)) < 0) {
			sm_snprintf(buf, sizeof(buf), "[%d] socket", tid);
			print_sys_error(buf);
			HDL_ERR(true);
		}

		/* bind here? */
		if (use_bind) {
			ret = clt_bind(sock);
			if (ret != SM_SUCCESS) {
				sm_snprintf(buf, sizeof(buf), "[%d] clt_bind()=%d", tid, ret);
				print_sys_error(buf);
				close(sock);
				HDL_ERR(true);
			}
		}

		ret = sm_io_open(&SmStThrIO, (void *) &sock, SM_IO_RDWR, &fp, NULL);
		if (ret != SM_SUCCESS) {
			sm_snprintf(buf, sizeof(buf), "[%d] sm_io_open()=%d", tid, ret);
			print_sys_error(buf);
			close(sock);
			HDL_ERR(true);
		}
		sm_io_clrblocking(fp);
		rmt_nfd = (st_netfd_t) f_cookie(*fp);

		r = request_timeout;
		ret = sm_io_setinfo(fp, SM_IO_WHAT_TIMEOUT, &r);
		if (ret != SM_SUCCESS) {
			sm_snprintf(buf, sizeof(buf), "[%d] set timeout()=%d", tid, ret);
			print_sys_error(buf);
			HDL_ERR(true);
		}
		ret = sm_io_setinfo(fp, SM_IO_DOUBLE, NULL);
		if (ret != SM_SUCCESS) {
			sm_snprintf(buf, sizeof(buf), "[%d] set double()=%d", tid, ret);
			print_sys_error(buf);
			HDL_ERR(true);
		}
#if MTA_NETINET6
/*
		if (debug > 2)
			sm_io_fprintf(smioerr, "sock_size=%d, fam=%d, port=%d, ip=%a\n", (int)SOCKSIZE(rmt_addr),(int)rmt_addr.sa.sa_family,(int)rmt_addr.sin6.sin6_port, &rmt_addr.sin6.sin6_addr);
*/
#endif
		if (st_connect(rmt_nfd, &rmt_addr.sa, SOCKSIZE(rmt_addr),
				SEC2USEC(request_timeout)) < 0)
		{
			sc_dnprintf(tid, buf, sizeof(buf), "connect(%a)", &rmt_addr);
			print_sys_error(buf);
			HDL_ERR(true);
		}
		++concurrent;
		incrcc = true;
		if (debug > 2)
			sc_dprintf(tid, "client: connected, ta=%u, conc=%d\n",
				ta_per_thrd, concurrent);

		if (hostinfo != NULL) {
			unsigned int prev;

			prev = nhostinfo;
			if (++nhostinfo >= nhostinfos)
				nhostinfo = 0;
			if (use_hi_perc) {
				unsigned long sesssions;

				sesssions = sess_cnt > 0 ? sess_cnt : 1;
				while (nhostinfo != prev) {
					if (hi_used[nhostinfo] * hi_total / sesssions < hi_perc[nhostinfo])
						break;
					if (++nhostinfo >= nhostinfos)
						nhostinfo = 0;
				}
			}
			hostinfo = hostinfos[nhostinfo];
			hi_used[nhostinfo]++;

			if (hostinfo[0] == 'R') {
				uchar *src;
				ipv4_T ipv4;

				ipv4 = random();
				src = (uchar *) &ipv4;
				n = sm_snprintf(buf, sizeof(buf), "X-CONNECT %u.%u.%u.%u\r\n",
						src[0], src[1], src[2], src[3]);
			}
			else {
				n = strlcpy(buf, "X-CONNECT ", sizeof(buf));
				n = strlcat(buf, hostinfo, sizeof(buf));
				n = strlcat(buf, "\r\n", sizeof(buf));
			}
			ret = sm_io_write(fp, (const uchar *)buf, strlen(buf), &b);
			if ((size_t) b != strlen(buf) || ret != SM_SUCCESS) {
				sc_dprintf(tid,
					"hostinfo write error n=%d, b=%d, ret=%#x\n"
					, strlen(buf), (int) b, ret);
				break;
			}
		}
		do {
			sm_str_clr(out);
			ret = sm_fgetline0(fp, out);

			if (debug > 3) {
					sc_dprintf(tid,
						"greet [len=%d, res=%#x]: ",
						sm_str_getlen(out), ret);
				sm_io_write(smioerr, sm_str_getdata(out),
					sm_str_getlen(out), &b);
				sm_io_flush(smioerr);
			}
			if (sm_is_error(ret)) {
				sc_dprintf(tid, "read greet ret=%#x\n", ret);
				HDL_ERR(true);
			}
		} while (!sm_is_error(ret) && sm_str_getlen(out) > 4 &&
			 sm_str_rd_elem(out, 3) == '-');
		if (stats && sc_sts != NULL) {
			t1 = st_utime();
			sc_stse->sces_rd[ST_SE_CONNECT] = t1 - t0;
			sc_stse->sces_wr[ST_SE_CONNECT] = 0;
		}
		if (!sm_is_error(ret)
		    && sm_str_getlen(out) > 2
		    && sm_str_rd_elem(out, 0) == '4'
		    && sm_str_rd_elem(out, 1) == '2'
		    && sm_str_rd_elem(out, 2) == '1')
		{
			if (showerror)
				sc_dprintf(tid,
					"greeting=error, text=%@T, sessions=%d\n",
					out, sessions);
			HDL_ERR(true);
		}
		if (sm_is_error(ret) || sm_str_getlen(out) <= 0
		    || sm_str_rd_elem(out, 0) != '2')
		{
			if (showerror)
				sc_dprintf(tid,
					"greeting=error, text=%@T, sessions=%d\n",
					out, sessions);
			HDL_ERR(true);
		}

		if (SCE_IS_ABORT_SE(codes[STAGE_CONNECT], sess_cnt))
			goto fail;

		sc_se.scse_srv_caps = 0;
		sc_se.scse_srv_max_sz_b = 0;
		sc_se.scse_seid[0] = '\0';

		if (cmdfile != NULL) {
			r = smtpcmds(fp, tid, out);
			if (sm_is_err(r))
				goto fail;
			cta = ta_per_thrd;
			goto snddata;	/* dirty hack... */
		}
		RANDOM_DROP;

  ehlo:
		n = strlcpy(buf, "EHLO ", sizeof(buf));
		n = strlcat(buf, ehloarg, sizeof(buf));
		n = strlcat(buf, "\r\n", sizeof(buf));
		if (stats && sc_sts != NULL) {
			pt0 = &sc_stse->sces_wr[ST_SE_EHLO];
			pt1 = &sc_stse->sces_rd[ST_SE_EHLO];
		}

		r = smtpcommand(buf, n, fp, out, tid, 0, STAGE_HELO, pt0, pt1, &sc_se, &sc_ta);
		if (SMTP_SSD == r) {
			HDL_ERR(true);
		}
		if (!SMTP_IS_OK(r)) {
			if (showerror)
				sc_dprintf(tid, "EHLO=error, text=%@T, sessions=%d\n",
					out, sessions);
			HDL_ERR(false);
		}

		if (SCE_IS_ABORT_SE(codes[STAGE_HELO], sess_cnt))
			goto fail;
		RANDOM_DROP;

#if MTA_USE_TLS
		if (try_starttls && tls_ctx != NULL &&
			SCSE_IS_CAP(sc_se.scse_srv_caps, SCSE_CAP_STARTTLS) && !tlsused)
		{
			ssize_t written;

			se_con = NULL;

			/* required by tls_verify_cert_cb()! */
			ret = tlsi_new(&se_tlsi, tls_ctx);
			if (sm_is_err(ret) || NULL == se_tlsi) {
				sc_dprintf(tid, "tlsi_new=%m\n", ret);
				goto notls;
			}

			se_con = SSL_new(tls_ctx->tls_ssl_ctx);
			if (NULL == se_con) {
				ret = sm_err_temp(SM_E_UNEXPECTED);
				goto notls;
			}

			r = SSL_set_ex_data(se_con, TLSctxidx, &sc_se);
			if (0 == r) {
				ret = sm_err_temp(SM_E_UNEXPECTED);
				goto notls;
			}

			r = SSL_set_ex_data(se_con, TLSsslidx, se_tlsi);
			if (0 == r) {
				ret = sm_err_temp(SM_E_UNEXPECTED);
				goto notls;
			}

			/* Hack... shouldn't use direct access but a function */
			se_tlsi->tlsi_log_field = "sc";
			se_tlsi->tlsi_log_value = "id";

			n = strlcpy(buf, "STARTTLS\r\n", sizeof(buf));
			r = smtpcommand(buf, n, fp, out, tid, 0, STAGE_TLS, pt0, pt1, &sc_se, &sc_ta);
			if (!SMTP_IS_OK(r)) {
				ret = sm_err_temp(SM_E_UNEXPECTED); /* ??? */
				goto notls;
			}

			SSL_set_connect_state(se_con);
			ret = tls_open(fp, se_con, tls_ctx, &fptls);
			if (sm_is_success(ret)) {
				/* HACK */
				fp = fptls;
				ret = do_tls_operation(fp,
					SSL_connect, NULL, NULL, NULL, 0, &written);
				if (sm_is_err(ret)) goto fail;
				r = 1;
				ret = sm_io_setinfo(fp, SM_IO_DOUBLE, &r);
				if (sm_is_err(ret)) goto notls;
				(void) tls_get_info(tls_ctx, se_con, TLS_F_CERT_RQST, se_tlsi);
				tlsused = true;
				goto ehlo;
			}
  notls:
			if (sm_is_err(ret)) {
				if (fp != fptls && se_con != NULL) {
					SSL_free(se_con);
					se_con = NULL;
				}
				goto fail;
			}
		}
#endif

		for (cta = 0;
		     cta < ta_per_thrd && myseq != seqfirst && sessions != 0;
		     cta++)
		{
			if (stoptime > 0 && st_time() > stoptime && sessions > 0) {
				sm_io_fprintf(smioerr, "stopped because of timelimit %ld\n", (long) stoptime);
				sessions = 0;
				break;
			}

			sc_ta.scta_rcpts_s = sc_ta.scta_rcpts_r = sc_ta.scta_rcpts_ok = 0;
			sc_ta.scta_rcpt_i = 0;
			sc_ta.scta_rcpt_r = sc_ta.scta_mail_r = SMTP_NONE;

			if (stats && sc_stt != NULL) {
				t0 = st_utime();
				sc_stta = &(sc_stt[cta]);
			}
			RANDOM_DROP;

			sm_memzero(sbuf, sizeof(sbuf));
			s = sbuf;
			if (usesize && datalen > 0)
				n = sm_snprintf(sbuf, sizeof(sbuf), " SIZE=%lu", datalen);
			if (mailarg != NULL) {
				/* ignore errors for now ... */
				strlcat(sbuf, " ", sizeof(sbuf));
				strlcat(sbuf, mailarg, sizeof(sbuf));
			}

			use_drr = try_drr && SCSE_IS_CAP(sc_se.scse_srv_caps, SCSE_CAP_DRR);
			use_rsad = try_rsad && SCSE_IS_CAP(sc_se.scse_srv_caps, SCSE_CAP_RSAD);

#define MAIL_ARG	(use_drr ? " DRR" : use_rsad ? " PRDR" : "")

			if (fromaddrs > 0)
				n = sm_snprintf(buf, sizeof(buf),
					"MAIL FROM:<%s>%s%s\r\n",
					from[mail_cnt % fromaddrs], s, MAIL_ARG);
			else if (myseq > 0)
				n = sm_snprintf(buf, sizeof(buf),
					"MAIL From:<nobody-%d-%d_%d@%s>%s\r\n",
					myseq, tid, postfix,
					fromdom[mail_cnt % fromdoms], MAIL_ARG);
			else if (uniqueaddrs) {
				n = sm_snprintf(buf, sizeof(buf),
					"MAIL from:<nobody-%lu@%s>%s%s\r\n",
					addrcnt,
					fromdom[mail_cnt % fromdoms], s, MAIL_ARG);
				++addrcnt;
			}
			else
				n = sm_snprintf(buf, sizeof(buf),
					"MAIL from:<nobody-%u-%d-%d@%s>%s%s\r\n",
					cta, tid, sessions,
					fromdom[mail_cnt % fromdoms], s, MAIL_ARG);
			if (mail_cnt > 0 && waits[STAGE_NEXT] > 0)
				sleep(waits[STAGE_NEXT]);
			++mail_cnt;
			if (stats && sc_sts != NULL) {
				pt0 = &sc_stta->scet_wr[ST_TA_MAIL];
				pt1 = &sc_stta->scet_rd[ST_TA_MAIL];
			}
			r = smtpcommand(buf, n, fp, out, tid, cta, STAGE_MAIL, pt0, pt1, &sc_se, &sc_ta);
			if (SMTP_SSD == r) {
				HDL_ERR(false);
			}
			if (!SMTP_IS_OK(r)) {
				if (showerror)
					sc_dprintfs(tid, sc_se.scse_seid,
						"MAIL=error, text=%@T, sessions=%d, ta=%u\n",
						out, sessions, cta);
				break;
			}
			if (SCE_IS_ABORT_SE(codes[STAGE_MAIL], ta_cnt))
				goto fail;
			if (SCE_IS_ABORT_TA(codes[STAGE_MAIL], ta_cnt))
				break;

#define RCPT_ARGS	(orcpt != NULL ? " " : ""), (orcpt != NULL ? orcpt : "")
			rok = 0;
			for (j = 0; j < rcpts; j++) {
				RANDOM_DROP;
				if (rcptaddrs > 0)
					n = sm_snprintf(buf, sizeof(buf),
						"RCPT TO:<%s>%s%s\r\n",
						rcpt[(rpct_per_ta ? j : rcpt_cnt) % rcptaddrs],
						RCPT_ARGS);
				else if (myseq > 0)
					n = sm_snprintf(buf, sizeof(buf),
						"RCPT To:<nobody-%d-%d-%d@%s>%s%s\r\n",
						myseq, j, tid,
						rcptdom[rcpt_cnt % rcptdoms], RCPT_ARGS);
				else if (uniqueaddrs) {
					n = sm_snprintf(buf, sizeof(buf),
						"RCPT to:<nobody-%lu@%s>%s%s\r\n",
						addrcnt,
						rcptdom[rcpt_cnt % rcptdoms], RCPT_ARGS);
					++addrcnt;
				}
				else
					n = sm_snprintf(buf, sizeof(buf),
						"RCPT to:<nobody-%u-%u-%d-%d@%s>\r\n",
						cta, j, tid, sessions,
						rcptdom[rcpt_cnt % rcptdoms]);
				++rcpt_cnt;
				if (stats && sc_sts != NULL && 0 == j) {
					pt0 = &sc_stta->scet_wr[ST_TA_RCPT];
					pt1 = &sc_stta->scet_rd[ST_TA_RCPT];
				}
				r = smtpcommand(buf, n, fp, out, tid, cta, STAGE_RCPT, pt0, pt1, &sc_se, &sc_ta);
				if (SMTP_SSD == r) {
					HDL_ERR(false);
				}
				if (SMTP_OK == r && !use_pipelining)
					++sc_ta.scta_rcpts_ok;

				/* pipelining... */
				if (use_pipelining && !SMTP_IS_OK(sc_ta.scta_mail_r)) {
					if (showerror)
						sc_dprintfs(tid, sc_se.scse_seid,
							"MAIL=error, text=%@T, sessions=%d, ta=%u\n",
							reply_str, sessions, cta);
					if (stoponerror)
						sessions = 0;
					break;
				}
				if (use_pipelining && !SMTP_IS_OK(sc_ta.scta_rcpt_r)) {
					if (showerror)
						sc_dprintfs(tid, sc_se.scse_seid,
							"RCPT=error, text=%@T, se=%d, ta=%u, rcpt=%u\n",
							reply_str, sessions, cta, sc_ta.scta_rcpt_i);
					if (stoponerror)
						sessions = 0;
					break;
				}

				if (!SMTP_IS_OK(r)) {
					if (showerror)
						sc_dprintfs(tid, sc_se.scse_seid,
							"RCPT=error, text=%@T, se=%d, ta=%u, rcpt=%u\n",
							out, sessions, cta, j);
					if (SMTP_IO_ERR(r))
						break;
					if (stoponerror)
						sessions = 0;
				}
				else
					++rok;
				if (SCE_IS_ABORT_SE(codes[STAGE_RCPT], ta_cnt))
					goto fail;
				if (SCE_IS_ABORT_TA(codes[STAGE_RCPT], ta_cnt))
					break;
				RANDOM_DROP;
			}
			if (SMTP_FATAL(r) || 0 == rok) {
				if (contiftaerror)
					goto next_ta;
				break;
			}
			if (SCE_IS_ABORT_TA(codes[STAGE_RCPT], ta_cnt))
				break;

  snddata:
			n = strlcpy(buf, "DATA\r\n", sizeof(buf));
			if (stats && sc_sts != NULL) {
				pt0 = &sc_stta->scet_wr[ST_TA_DATA];
				pt1 = &sc_stta->scet_rd[ST_TA_DATA];
			}
			r = smtpcommand(buf, n, fp, out, tid, cta, STAGE_DATA, pt0, pt1, &sc_se, &sc_ta);
			if (SMTP_SSD == r) {
				HDL_ERR(false);
			}

			/* pipelining... */
			if (use_pipelining && !SMTP_IS_OK(sc_ta.scta_mail_r)) {
				if (showerror)
					sc_dprintfs(tid, sc_se.scse_seid,
						"MAIL=error, text=%@T, sessions=%d, ta=%u\n",
						reply_str, sessions, cta);
				if (stoponerror)
					sessions = 0;
				break;
			}
			if (use_pipelining && !SMTP_IS_OK(sc_ta.scta_rcpt_r)) {
				if (showerror)
					sc_dprintfs(tid, sc_se.scse_seid,
						"RCPT=error, text=%@T, se=%d, ta=%u, rcpt=%u\n",
						reply_str, sessions, cta, sc_ta.scta_rcpt_i);
				if (stoponerror) {
					sessions = 0;
					break;
				}
				if (sc_ta.scta_rcpts_ok <= 0) {
					if (showerror)
						sc_dprintfs(tid, sc_se.scse_seid,
							"DATA=error, rcpts_ok=%d\n",
							sc_ta.scta_rcpts_ok);
					if (contiftaerror)
						goto next_ta;
					break;
				}
			}

			if (!SMTP_IS_CONT(r)) {
				HDL_ERR2("DATA", r);
			}
			if (waitdata > 0)
				st_sleep(waitdata);
			if (SCE_IS_ABORT_SE(codes[STAGE_DATA], ta_cnt))
				goto fail;
			if (SCE_IS_ABORT_TA(codes[STAGE_DATA], ta_cnt))
				break;
			RANDOM_DROP;

			if (myseq > 0) {
				n = sm_snprintf(buf, sizeof(buf),
					"From: me+%d\r\nTo: you+%d\r\nMessage-Id: <%d@smtpc2.domain>\r\nSubject: test+%d\r\n\r\n%d\r\n",
					myseq, myseq, myseq, myseq, myseq);
				ret = sm_io_write(fp, (unsigned char *) buf,
						strlen(buf), &b);
				if ((size_t) b != strlen(buf) || ret != SM_SUCCESS) {
					sc_dprintfs(tid, sc_se.scse_seid,
						"data write error, n=%d, b=%d, ret=%#x\n",
						strlen(buf), (int) b, ret);
					break;
				}
				if (datalen > 0 || datafile != NULL) {
					r = smtpdata(datalen, fp, tid, cta);
					if (!SMTP_IS_OK(r))
						break;
				}
				n = strlcpy(buf, "\r\n.\r\n", sizeof(buf));
			}
			else if (0 == datalen && NULL == datafile) {
				n = strlcpy(buf,
					"From: me\r\nTo: you\r\nSubject: test\r\n\r\nbody\r\n.\r\n",
					sizeof(buf));
			}
			else {
				r = smtpdata(datalen, fp, tid, cta);
				if (!SMTP_IS_OK(r))
					break;
				n = strlcpy(buf, "\r\n.\r\n", sizeof(buf));
			}
			if (SCE_IS_ABORT_SE(codes[STAGE_DOT], ta_cnt))
				goto fail;
			if (SCE_IS_ABORT_TA(codes[STAGE_DOT], ta_cnt))
				break;
			RANDOM_DROP;

			if (stats && sc_sts != NULL) {
				pt0 = &(sc_stta->scet_wr[ST_TA_DOT]);
				pt1 = &(sc_stta->scet_rd[ST_TA_DOT]);
			}
			r = smtpcommand(buf, n, fp, out, tid, cta, STAGE_DOT, pt0, pt1, &sc_se, &sc_ta);
			if (SMTP_SSD == r) {
				HDL_ERR(false);
			}
			if (!SMTP_IS_OK(r)) {
				HDL_ERR2("DOT", r);
			}

  next_ta:
			/* another transaction? */
			if (cta < ta_per_thrd - 1) {
				myseq = sequence;
				if (sequence > seqfirst)
					--sequence;
				if (rset) {
					n = strlcpy(buf, "RSET\r\n", sizeof(buf));
					if (stats && sc_sts != NULL) {
						pt0 = &sc_stta->scet_wr[ST_TA_RSET];
						pt1 = &sc_stta->scet_rd[ST_TA_RSET];
					}
					r = smtpcommand(buf, n, fp, out, tid, cta, -1, pt0, pt1, &sc_se, &sc_ta);
					if (!SMTP_IS_OK(r)) {
						/* classify error: need to start new session? */
						if (SMTP_IO_ERR(r) && contiftaerror > 1)
							goto next_se;
						if (!contiftaerror)
							break;
					}
				}
			}

			if (stats && sc_stt != NULL) {
				t1 = st_utime();
				sc_stta->scet_ta = t1 - t0;
			}

			if (SMTP_IS_OK(r))
				++ta_cnt;
			if (count > 0 && (ta_cnt % count) == 0) {
				time_t lastNmail, elapsedN;
				time_t elapsed;
				long tr;

				/* sm_io_fprintf(smioerr, "%ld\r", ta_cnt); */
				lastNmail = st_time();
				elapsedN = lastNmail - firstNmail;
				firstNmail = lastNmail;
				tr = 0;
				if (firstmail > 0) {
					elapsed = lastNmail - firstmail;
					if (elapsed > 0)
						tr = (ta_cnt - startup_act) / elapsed;
				}
				else
					elapsed = 0;
				if (idisplay)
					sm_io_fprintf(smioerr,
						"%6u  %8ld %8ld  %6ld   %6ld  %4d  %4d\r"
						, ta_cnt
						, (long) elapsedN
						, (long) elapsed
						, (elapsedN > 0) ? (long) (count / elapsedN) : -1L
						, tr
						, busy, concurrent
						);
				else
					sm_io_fprintf(smioerr,
						"msg=%6u, t_i=%2ld, t_t=%4ld, msg/s=%4ld, t-msg/s=%4ld, busy=%3d, conc=%3d\n"
						, ta_cnt
						, (long) elapsedN
						, (long) elapsed
						, (elapsedN > 0) ? (long) (count / elapsedN) : -1L
						, tr
						, busy, concurrent
						);

				if (ta_cnt + teardown >= ta_total && !total_rate_set) {
					total_rate = tr;
					total_rate_set = true;
					if (teardown > 0 && count >= 10)
						count /= 10;
				}

			}
			if (ta_total > 0 && ta_cnt >= ta_total) {
				sessions = 0;
				if (!total_rate_set) {
					time_t lastNmail, elapsedN;
					time_t elapsed;
					long tr;

					lastNmail = st_time();
					elapsedN = lastNmail - firstNmail;
					firstNmail = lastNmail;
					tr = 0;
					if (firstmail > 0) {
						elapsed = lastNmail - firstmail;
						if (elapsed > 0)
							tr = (ta_cnt - startup_act) / elapsed;
					}
					if (ta_cnt + teardown >= ta_total && !total_rate_set) {
						total_rate = tr;
						total_rate_set = true;
						if (teardown > 0 && count >= 10)
							count /= 10;
					}
				}
				break;
			}
		}
		if (stats && sc_sts != NULL) {
			pt0 = &sc_stse->sces_wr[ST_SE_QUIT];
			pt1 = &sc_stse->sces_rd[ST_SE_QUIT];
		}
		n = strlcpy(buf, "QUIT\r\n", sizeof(buf));
		r = smtpcommand(buf, n, fp, out, tid, cta, STAGE_QUIT, pt0, pt1, &sc_se, &sc_ta);
		/*
		if (!SMTP_IS_OK(r))
			;
		*/

  next_se:
		if (fp != NULL) {
			sm_io_close(fp, SM_IO_CF_NONE);
			fp = NULL;
		}
		if (incrcc) {
			--concurrent;
			incrcc = false;
		}
		if (stats && sc_sts != NULL)
			sc_stse->sces_se = st_utime() - ts0;
	}

	if (stoponerror)
		sessions = 0;
  fail:
	if (fp != NULL) {
		sm_io_close(fp, SM_IO_CF_NONE);
		fp = NULL;
	}
	if (incrcc) {
		--concurrent;
		incrcc = false;
	}

  done:
	if (debug > 0)
		sc_dprintf(tid,
			"where=done, sess_cnt=%u, sessions=%u\n",
			sess_cnt, sessions);
#if MTA_USE_TLS
	(void) tlsi_free(se_tlsi);
#endif
	SM_STR_FREE(out);
	--busy;
	return NULL;
}

static void
print_sys_error(const char *msg)
{
	sm_io_fprintf(smioerr, "%s: %s: %s\n", prog, msg, strerror(errno));
}
